# flake8: noqa F401 (imported but unused)
import abc
import unittest

import numpy as np
import six

from kafe2.config import kc
from kafe2.core.minimizers import AVAILABLE_MINIMIZERS
from kafe2.fit import IndexedFit
from kafe2.fit.indexed.cost import IndexedCostFunction_Chi2
from kafe2.test.fit.test_fit import AbstractTestFit


def simple_chi2(data, model):
    return np.sum((data - model) ** 2)


def simple_chi2_explicit_model_name(data, simple_indexed_model):
    return np.sum((data - simple_indexed_model) ** 2)


def simple_indexed_model(a=1.1, b=2.2, c=3.3):
    x = np.arange(10)
    return a * x**2 + b * x + c


def line_indexed_model(a=3.0, b=0.0):
    x = np.arange(10)
    return a * x + b


def analytic_solution(des_mat, cov_mat_inv, data):
    return np.squeeze(np.asarray(np.linalg.inv(des_mat.T.dot(cov_mat_inv).dot(des_mat)).dot(des_mat.T).dot(cov_mat_inv).dot(data)))


class TestIndexedFitBasicInterface(AbstractTestFit, unittest.TestCase):
    MINIMIZER = "scipy"

    def setUp(self):
        self._n_points = 10

        # "jitter" for data smearing
        self._jitter = np.array(
            [
                -0.3193475,
                -1.2404198,
                -1.4906926,
                -0.78832446,
                -1.7638106,
                0.36664261,
                0.49433821,
                0.0719646,
                1.95670326,
                0.31200215,
            ]
        )
        assert len(self._jitter) == self._n_points

        # reference initial values
        self._ref_initial_pars = np.array([1.1, 2.2, 3.3])
        self._ref_initial_model = simple_indexed_model(*self._ref_initial_pars)

        # fit data
        self._ref_data = self._ref_initial_model + self._jitter

        self._linear_design_matrix = np.array([np.arange(10) ** 2, np.arange(10), np.ones(10)]).T

        # pre-fit cost value
        self._ref_initial_cost = simple_chi2(
            data=self._ref_data,
            model=self._ref_initial_model,
        )

        # reference matrices/errors
        self._ref_error = np.ones(self._n_points)
        self._ref_matrix_eye = np.eye(self._n_points)

        # reference fit result values
        # self._nominal_fit_result_pars = (1.1351433, 2.13736919, 2.33346549)
        self._nominal_fit_result_pars = analytic_solution(
            self._linear_design_matrix,
            np.linalg.inv(self._ref_matrix_eye),
            self._ref_data,
        )

        self._nominal_fit_result_model = simple_indexed_model(*self._nominal_fit_result_pars)
        self._nominal_fit_result_cost = simple_chi2(
            data=self._ref_data,
            model=self._nominal_fit_result_model,
        )

        # helper dict with all reference property values
        self._ref_prop_dict = dict(
            did_fit=False,
            model_count=1,
            parameter_values=self._ref_initial_pars,
            parameter_names=("a", "b", "c"),
            cost_function_value=self._ref_initial_cost,
            data=self._ref_data,
            data_error=self._ref_error,
            data_cov_mat=self._ref_matrix_eye,
            data_cov_mat_inverse=self._ref_matrix_eye,
            data_cor_mat=self._ref_matrix_eye,
            model=self._ref_initial_model,
            model_error=self._ref_error * 0,
            model_cov_mat=self._ref_matrix_eye * 0,
            model_cov_mat_inverse=None,
            model_cor_mat=self._ref_matrix_eye * np.nan,
            total_error=self._ref_error,
            total_cov_mat=self._ref_matrix_eye,
            total_cov_mat_inverse=self._ref_matrix_eye,
            total_cor_mat=self._ref_matrix_eye,
        )

    def _get_fit(self, model_function=None, cost_function=None, errors=None, dynamic_error_algorithm=None):
        """convenience"""
        model_function = model_function or simple_indexed_model
        # TODO: fix default
        cost_function = cost_function or IndexedCostFunction_Chi2(errors_to_use="covariance")
        errors = errors or [dict(err_val=1.0)]
        dynamic_error_algorithm = dynamic_error_algorithm or "nonlinear"

        _fit = IndexedFit(
            data=self._ref_data,
            model_function=model_function,
            cost_function=cost_function,
            minimizer=self.MINIMIZER,
            dynamic_error_algorithm=dynamic_error_algorithm,
        )
        for _err in errors:
            _fit.add_error(**_err)

        return _fit

    def _get_test_fits(self):
        return {
            "default": self._get_fit(),
            "explicit": self._get_fit(cost_function=simple_chi2),
            "explicit_model": self._get_fit(cost_function=simple_chi2_explicit_model_name),
            "relative_errors_data": self._get_fit(errors=[dict(err_val=1.0 / self._ref_data, relative=True, reference="data")]),
        }

    def test_initial_state(self):
        self.run_test_for_all_fits(self._ref_prop_dict)

    def test_fit_results(self):
        self.run_test_for_all_fits(
            dict(
                self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                model=self._nominal_fit_result_model,
                did_fit=True,
                cost_function_value=self._nominal_fit_result_cost,
            ),
            call_before_fit=lambda f: f.do_fit(),
            rtol=1e-2,
        )

    def test_set_all_parameter_values(self):
        self.run_test_for_all_fits(
            dict(
                self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                data=self._ref_data,
                model=self._nominal_fit_result_model,
                parameter_names=("a", "b", "c"),
                did_fit=False,
                model_count=1,
                cost_function_value=self._nominal_fit_result_cost,
            ),
            # set parameters to their final values by hand
            call_before_fit=lambda f: f.set_all_parameter_values(self._nominal_fit_result_pars),
        )

    def test_set_all_parameter_values_wrong_number_raise(self):
        with self.assertRaises(ValueError):
            self._get_fit().set_all_parameter_values((1,))
        with self.assertRaises(ValueError):
            self._get_fit().set_all_parameter_values((1, 2, 3, 4, 5))

    def test_parameter_defaults(self):
        def dummy_model(a, b, c):
            return np.arange(10)

        _fit = self._get_fit(model_function=dummy_model)

        self._assert_fit_properties(
            _fit,
            dict(
                parameter_values=np.array([1, 1, 1]),
            ),
        )

    def test_parameter_partial_defaults(self):
        def dummy_model(a, b, c=3.3):
            return np.arange(10)

        _fit = self._get_fit(model_function=dummy_model)

        self._assert_fit_properties(
            _fit,
            dict(
                parameter_values=np.array([1, 1, 3.3]),
            ),
        )

    def test_update_data(self):
        _fit = self._get_fit()

        _fit.data = self._ref_data * 2
        _fit.add_error(err_val=1.0)

        self._assert_fit_properties(
            _fit,
            dict(
                data=self._ref_data * 2,
            ),
        )

        _fit.do_fit()
        _new_estimates = np.array(self._nominal_fit_result_pars) * 2

        self._assert_fit_properties(
            _fit,
            dict(
                data=self._ref_data * 2,
                parameter_values=_new_estimates,
            ),
            rtol=1e-2,
        )

    def test_update_data_length_mismatch_raise(self):
        # TODO: update when different length limitation is fixed
        _fit = self._get_fit()
        with self.assertRaises(ValueError):
            _fit.data = np.arange(self._n_points + 1)

    def test_reserved_parameter_names_raise(self):
        def dummy_model(data):
            pass

        with self.assertRaises(ValueError) as _exc:
            IndexedFit(data=self._ref_data, model_function=dummy_model, minimizer=self.MINIMIZER)

        self.assertIn("reserved", _exc.exception.args[0])
        self.assertIn("data", _exc.exception.args[0])

    def dummy_model_no_pars_raise(self):
        def dummy_model():
            pass

        with self.assertRaises(ValueError) as _exc:
            IndexedFit(data=self._ref_data, model_function=dummy_model, minimizer=self.MINIMIZER)

        self.assertIn("needs at least one parameter", _exc.exception.args[0])

    def dummy_model_varargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, *varargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            IndexedFit(data=self._ref_data, model_function=dummy_model, minimizer=self.MINIMIZER)

        self.assertIn("variable", _exc.exception.args[0])
        self.assertIn("varargs", _exc.exception.args[0])

    def dummy_model_varkwargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, **varkwargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            IndexedFit(data=self._ref_data, model_function=dummy_model, minimizer=self.MINIMIZER)

        self.assertIn("variable", _exc.exception.args[0])
        self.assertIn("varkwargs", _exc.exception.args[0])

    def dummy_model_varargs_varkwargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, *varargs, **varkwargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            IndexedFit(data=self._ref_data, model_function=dummy_model, minimizer=self.MINIMIZER)

        self.assertIn("variable", _exc.exception.args[0])
        self.assertIn("varargs", _exc.exception.args[0])
        # TODO: enable when implemented
        # self.assertIn('varkwargs', _exc.exception.args[0])

    def test_report_before_fit(self):
        # TODO: check report content
        _buffer = six.StringIO()
        _fit = self._get_fit()
        _fit.report(output_stream=_buffer)
        self.assertNotEqual(_buffer.getvalue(), "")

    def test_report_after_fit(self):
        # TODO: check report content
        _buffer = six.StringIO()
        _fit = self._get_fit()
        _fit.do_fit()
        _fit.report(output_stream=_buffer)
        self.assertNotEqual(_buffer.getvalue(), "")

    def test_relative_model_error_nonlinear(self):
        _fit_data_err = self._get_fit(
            errors=[
                dict(err_val=1.0, relative=False, reference="data"),
                dict(err_val=0.1, relative=True, reference="data"),
            ]
        )
        _fit_model_err = self._get_fit(
            errors=[
                dict(err_val=1.0, relative=False, reference="data"),
                dict(err_val=0.1, relative=True, reference="model"),
            ],
            dynamic_error_algorithm="nonlinear",
        )
        _fit_data_err.do_fit()
        _fit_model_err.do_fit()
        self._assert_fit_results_equal(_fit_data_err, _fit_model_err, rtol=1e-2, atol=2e-2)

    def test_relative_model_error_iterative(self):
        _fit_data_err = self._get_fit(
            errors=[
                dict(err_val=1.0, relative=False, reference="data"),
                dict(err_val=0.1, relative=True, reference="data"),
            ]
        )
        _fit_model_err = self._get_fit(
            errors=[
                dict(err_val=1.0, relative=False, reference="data"),
                dict(err_val=0.1, relative=True, reference="model"),
            ],
            dynamic_error_algorithm="iterative",
        )
        _fit_data_err.do_fit()
        _fit_model_err.do_fit()
        self._assert_fit_results_equal(_fit_data_err, _fit_model_err, rtol=1e-2)

    def test_fit_scalar_model_function(self):
        def _model_function(a):
            return a

        _fit = IndexedFit([1, 2], _model_function)
        _fit.add_error(1.0)
        _fit.do_fit()
        self.assertAlmostEqual(_fit.parameter_values[0], 1.5)
        _fit.report()


class TestIndexedFitWithSimpleErrors(AbstractTestFit, unittest.TestCase):
    MINIMIZER = "scipy"

    def setUp(self):
        six.get_unbound_function(TestIndexedFitBasicInterface.setUp)(self)

    def _get_fit(self, errors=None):
        """convenience"""

        errors = errors or [dict(err_val=1.0)]

        _fit = IndexedFit(
            data=self._ref_data,
            model_function=simple_indexed_model,
            cost_function=IndexedCostFunction_Chi2(errors_to_use="covariance"),
            minimizer=self.MINIMIZER,
        )

        for _err in errors:
            _fit.add_error(**_err)

        return _fit

    def _get_test_fits(self):
        return {
            "default": self._get_fit(),
            "two_errors": self._get_fit(errors=[dict(err_val=1.0 / np.sqrt(2)), dict(err_val=1.0 / np.sqrt(2))]),
            "named_errors": self._get_fit(
                errors=[
                    dict(
                        err_val=1.0 / np.sqrt(2),
                        name="MyDataError",
                        correlation=0,
                        relative=False,
                        reference="data",
                    ),
                    dict(
                        err_val=1.0 / np.sqrt(2),
                        name="MyModelError",
                        correlation=0,
                        relative=False,
                        reference="model",
                    ),
                ]
            ),
        }

    def test_initial_state(self):
        self.run_test_for_all_fits(
            self._ref_prop_dict,
            fit_names=["default", "two_errors"],
        )

    def test_fit_results(self):
        self.run_test_for_all_fits(
            dict(
                self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                model=self._nominal_fit_result_model,
                did_fit=True,
                cost_function_value=self._nominal_fit_result_cost,
            ),
            fit_names=["default", "two_errors"],
            call_before_fit=lambda f: f.do_fit(),
            rtol=1e-2,
        )

    def test_get_matching_error_all(self):
        _fit = self._get_test_fits()["named_errors"]
        for _mc in (None, dict()):
            _errs = _fit.get_matching_errors(matching_criteria=_mc)
            self.assertEqual(len(_errs), 2)
            self.assertIs(_fit.data_container._error_dicts["MyDataError"]["err"], _errs["MyDataError"])
            self.assertIs(_fit._param_model._error_dicts["MyModelError"]["err"], _errs["MyModelError"])

    def test_get_matching_error_name(self):
        _fit = self._get_test_fits()["named_errors"]
        _errs = _fit.get_matching_errors(matching_criteria=dict(name="MyDataError"))
        self.assertEqual(len(_errs), 1)
        self.assertIs(_fit.data_container._error_dicts["MyDataError"]["err"], _errs["MyDataError"])

    def test_get_matching_error_type_simple(self):
        _fit = self._get_test_fits()["named_errors"]
        _errs = _fit.get_matching_errors(matching_criteria=dict(type="simple"))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts["MyDataError"]["err"], _errs["MyDataError"])
        self.assertIs(_fit._param_model._error_dicts["MyModelError"]["err"], _errs["MyModelError"])

    def test_get_matching_error_type_matrix(self):
        _fit = self._get_test_fits()["named_errors"]
        _errs = _fit.get_matching_errors(matching_criteria=dict(type="matrix"))
        self.assertEqual(len(_errs), 0)

    def test_get_matching_error_uncorrelated(self):
        _fit = self._get_test_fits()["named_errors"]
        _errs = _fit.get_matching_errors(matching_criteria=dict(correlated=True))
        self.assertEqual(len(_errs), 0)

    def test_get_matching_error_correlated(self):
        _fit = self._get_test_fits()["named_errors"]
        _errs = _fit.get_matching_errors(matching_criteria=dict(correlated=False))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts["MyDataError"]["err"], _errs["MyDataError"])
        self.assertIs(_fit._param_model._error_dicts["MyModelError"]["err"], _errs["MyModelError"])

    def test_get_matching_error_reference_data(self):
        _fit = self._get_test_fits()["named_errors"]
        _errs = _fit.get_matching_errors(matching_criteria=dict(reference="data"))
        self.assertEqual(len(_errs), 1)
        self.assertIs(_fit.data_container._error_dicts["MyDataError"]["err"], _errs["MyDataError"])

    def test_get_matching_error_reference_model(self):
        _fit = self._get_test_fits()["named_errors"]
        _errs = _fit.get_matching_errors(matching_criteria=dict(reference="model"))
        self.assertEqual(len(_errs), 1)
        self.assertIs(_fit._param_model._error_dicts["MyModelError"]["err"], _errs["MyModelError"])


class TestIndexedFitWithMatrixErrors(AbstractTestFit, unittest.TestCase):
    MINIMIZER = "scipy"

    def setUp(self):
        six.get_unbound_function(TestIndexedFitBasicInterface.setUp)(self)

    def _get_fit(self, errors=None):
        """convenience"""

        _fit = IndexedFit(
            data=self._ref_data,
            model_function=simple_indexed_model,
            cost_function=IndexedCostFunction_Chi2(errors_to_use="covariance"),
            minimizer=self.MINIMIZER,
        )

        if errors is None:
            _fit.add_matrix_error(err_matrix=np.eye(self._n_points), matrix_type="cov")
        else:
            for _err in errors:
                if "err_matrix" in _err:
                    _fit.add_matrix_error(**_err)
                else:
                    _fit.add_error(**_err)

        return _fit

    def _get_test_fits(self):
        return {
            "default": self._get_fit(),
            "cor_matrix_and_error_vector": self._get_fit(errors=[dict(err_matrix=np.eye(self._n_points), matrix_type="cor", err_val=1.0)]),
            "two_matrix_errors": self._get_fit(
                errors=[
                    dict(err_matrix=np.eye(self._n_points) / 2, matrix_type="cov"),
                    dict(err_matrix=np.eye(self._n_points) / 2, matrix_type="cov"),
                ]
            ),
            "one_matrix_one_simple_error": self._get_fit(
                errors=[
                    dict(err_matrix=np.eye(self._n_points) / 2, matrix_type="cov"),
                    dict(err_val=1 / np.sqrt(2)),
                ]
            ),
            "named_errors": self._get_fit(
                errors=[
                    dict(
                        err_val=1.0 / np.sqrt(2),
                        name="MySimpleDataError",
                        correlation=0,
                        relative=False,
                        reference="data",
                    ),
                    dict(
                        err_val=1.0 / np.sqrt(2),
                        name="MySimpleModelError",
                        correlation=0,
                        relative=False,
                        reference="model",
                    ),
                    dict(
                        err_matrix=np.eye(self._n_points) / 2,
                        matrix_type="cov",
                        name="MyMatrixDataError",
                        relative=False,
                        reference="data",
                    ),
                    dict(
                        err_matrix=np.eye(self._n_points) / 2,
                        matrix_type="cov",
                        name="MyMatrixModelError",
                        relative=False,
                        reference="model",
                    ),
                ]
            ),
        }

    def test_initial_state(self):
        self.run_test_for_all_fits(
            self._ref_prop_dict,
            fit_names=[
                "default",
                "cor_matrix_and_error_vector",
                "one_matrix_one_simple_error",
                "two_matrix_errors",
            ],
        )

    def test_fit_results(self):
        self.run_test_for_all_fits(
            dict(
                self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                model=self._nominal_fit_result_model,
                did_fit=True,
                cost_function_value=self._nominal_fit_result_cost,
            ),
            fit_names=[
                "default",
                "cor_matrix_and_error_vector",
                "one_matrix_one_simple_error",
                "two_matrix_errors",
            ],
            call_before_fit=lambda f: f.do_fit(),
            rtol=1e-2,
        )

    def test_get_matching_error_all(self):
        _fit = self._get_test_fits()["named_errors"]
        for _mc in (None, dict()):
            _errs = _fit.get_matching_errors(matching_criteria=_mc)
            self.assertEqual(len(_errs), 4)
            self.assertIs(
                _fit.data_container._error_dicts["MySimpleDataError"]["err"],
                _errs["MySimpleDataError"],
            )
            self.assertIs(
                _fit._param_model._error_dicts["MySimpleModelError"]["err"],
                _errs["MySimpleModelError"],
            )
            self.assertIs(
                _fit.data_container._error_dicts["MyMatrixDataError"]["err"],
                _errs["MyMatrixDataError"],
            )
            self.assertIs(
                _fit._param_model._error_dicts["MyMatrixModelError"]["err"],
                _errs["MyMatrixModelError"],
            )

    def test_get_matching_error_name(self):
        _fit = self._get_test_fits()["named_errors"]
        _errs = _fit.get_matching_errors(matching_criteria=dict(name="MySimpleDataError"))
        self.assertEqual(len(_errs), 1)
        self.assertIs(_fit.data_container._error_dicts["MySimpleDataError"]["err"], _errs["MySimpleDataError"])

    def test_get_matching_error_type_simple(self):
        _fit = self._get_test_fits()["named_errors"]
        _errs = _fit.get_matching_errors(matching_criteria=dict(type="simple"))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts["MySimpleDataError"]["err"], _errs["MySimpleDataError"])
        self.assertIs(_fit._param_model._error_dicts["MySimpleModelError"]["err"], _errs["MySimpleModelError"])

    def test_get_matching_error_type_matrix(self):
        _fit = self._get_test_fits()["named_errors"]
        _errs = _fit.get_matching_errors(matching_criteria=dict(type="matrix"))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts["MyMatrixDataError"]["err"], _errs["MyMatrixDataError"])
        self.assertIs(_fit._param_model._error_dicts["MyMatrixModelError"]["err"], _errs["MyMatrixModelError"])

    def test_get_matching_error_uncorrelated(self):
        _fit = self._get_test_fits()["named_errors"]
        _errs = _fit.get_matching_errors(matching_criteria=dict(correlated=True))
        self.assertEqual(len(_errs), 0)

    def test_get_matching_error_correlated(self):
        _fit = self._get_test_fits()["named_errors"]
        _errs = _fit.get_matching_errors(matching_criteria=dict(correlated=False))
        # NOTE: passing 'correlated' only matches 'matrix' errors, irrespective of 'True'/'False' value passed
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts["MySimpleDataError"]["err"], _errs["MySimpleDataError"])
        self.assertIs(_fit._param_model._error_dicts["MySimpleModelError"]["err"], _errs["MySimpleModelError"])

    def test_get_matching_error_reference_data(self):
        _fit = self._get_test_fits()["named_errors"]
        _errs = _fit.get_matching_errors(matching_criteria=dict(reference="data"))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts["MySimpleDataError"]["err"], _errs["MySimpleDataError"])
        self.assertIs(_fit.data_container._error_dicts["MyMatrixDataError"]["err"], _errs["MyMatrixDataError"])

    def test_get_matching_error_reference_model(self):
        _fit = self._get_test_fits()["named_errors"]
        _errs = _fit.get_matching_errors(matching_criteria=dict(reference="model"))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit._param_model._error_dicts["MySimpleModelError"]["err"], _errs["MySimpleModelError"])
        self.assertIs(_fit._param_model._error_dicts["MyMatrixModelError"]["err"], _errs["MyMatrixModelError"])



import re

import numpy as np
import yaml

from ....core.error import MatrixGaussianError, SimpleGaussianError
from ...xy import XYContainer

__all__ = [
    "add_error_to_container",
    "write_errors_to_yaml",
    "process_error_sources",
    "MatrixYamlDumper",
    "MatrixYamlLoader",
]


_yaml_error_section_for_axis = {0: "x_errors", 1: "y_errors", None: "errors"}


def add_error_to_container(err_type, container_obj, **kwargs):
    # TODO: check kwargs explicitly
    if err_type == "simple":
        container_obj.add_error(**kwargs)
    elif err_type == "matrix":
        container_obj.add_matrix_error(**kwargs)
    else:
        raise TypeError("Unknown error type '{}'. " "Valid: {}".format(err_type, ("simple", "matrix")))
    return container_obj


def write_errors_to_yaml(container, yaml_doc):
    # TODO: create public error retrieval interface
    for _err_name, _err_dict in container._error_dicts.items():
        _err_obj = _err_dict["err"]
        _err_axis = _err_dict.get("axis", None)

        # get the relevant error section for the current axis, creating it if necessary
        _yaml_section = yaml_doc.setdefault(_yaml_error_section_for_axis[_err_axis], [])

        # -- check for relative errors
        _is_relative = _err_obj.relative
        if _is_relative:
            _err_val = _err_obj.error_rel
        else:
            _err_val = _err_obj.error

        # -- collapse identical errors to one float
        if np.allclose(_err_val[0], _err_val):
            _err_val = float(_err_val[0])
        else:
            _err_val = _err_val.tolist()

        # -- handle different error types
        # TODO shouldn't each error be wrapped inside an 'error' namespace?
        if _err_obj.__class__ is SimpleGaussianError:
            _yaml_section.append(
                dict(
                    name=_err_name,
                    type="simple",
                    error_value=_err_val,
                    relative=_is_relative,
                    correlation_coefficient=_err_obj._corr_coeff,
                    # TODO: public interface for _corr_coeff!
                )
            )
        elif _err_obj.__class__ is MatrixGaussianError:
            _mtype = _err_obj._matrix_type_at_construction  # TODO: public interface!
            _yaml_section.append(
                dict(
                    name=_err_name,
                    type="matrix",
                    matrix_type=_mtype,
                    relative=_is_relative,
                )
            )
            if _mtype == "covariance":
                if _is_relative:
                    _yaml_section[-1]["matrix"] = _err_obj.cov_mat_rel  # .tolist()
                else:
                    _yaml_section[-1]["matrix"] = _err_obj.cov_mat  # .tolist()
            elif _mtype == "correlation":
                _yaml_section[-1]["matrix"] = _err_obj.cor_mat  # .tolist()
                _yaml_section[-1]["error_value"] = _err_val
            else:
                raise TypeError("Unknown error matrix type '{}'. " "Valid: 'correlation' or 'covariance'.")
        else:
            raise TypeError("No representation for error type {} " "implemented!".format(type(_err_obj)))

    return yaml_doc


def process_error_sources(container_obj, yaml_doc):
    # -- process error sources
    # errors can be specified as a single float, a list of floats, or a kafe2 error object
    # lists of the above are also valid, if the error object is not a list
    if isinstance(container_obj, XYContainer):  # also applies for XYParamModel
        _xerrs = yaml_doc.pop("x_errors", [])
        if isinstance(_xerrs, (int, float, str)):
            _xerrs = [_xerrs] * container_obj.size
        if len(_xerrs) > 0 and isinstance(_xerrs[0], (int, float, str)):
            _xerrs = [_xerrs]
        _yerrs = yaml_doc.pop("y_errors", [])
        if not isinstance(_yerrs, list):
            _yerrs = [_yerrs] * container_obj.size
        if len(_yerrs) > 0 and isinstance(_yerrs[0], (int, float, str)):
            _yerrs = [_yerrs]
        _errs = _xerrs + _yerrs
        _axes = [0] * len(_xerrs) + [1] * len(_yerrs)  # 0 for 'x', 1 for 'y'
    else:
        _errs = yaml_doc.pop("errors", [])
        if not isinstance(_errs, list):
            _errs = [_errs] * container_obj.size
        if len(_errs) > 0 and isinstance(_errs[0], float):
            _errs = [_errs]
        _axes = [None] * len(_errs)

    # add error sources, if any
    for _err, _axis in zip(_errs, _axes):
        # If error is a list, check for mixture of relative/absolute errors.
        # Add it as a simple error and don't try to interpret it as a kafe2 error object
        if isinstance(_err, list):
            _rel = np.zeros(len(_err))
            _abs = np.zeros(len(_err))
            for i, _val in enumerate(_err):
                if isinstance(_val, str) and _val.endswith("%"):
                    try:
                        _rel_err_percent = float(_val[:-1])
                    except ValueError:
                        raise ValueError("Cannot convert string to relative error: %s" % _err)
                    _rel[i] = _rel_err_percent
                else:
                    try:
                        _abs_err = float(_val)
                    except ValueError:
                        raise ValueError("Cannot convert {} to error value.".format(_val))
                    _abs[i] = _val
            if _axis is not None:
                container_obj = add_error_to_container("simple", container_obj, err_val=0.01 * _rel, relative=True, axis=_axis)
                container_obj = add_error_to_container("simple", container_obj, err_val=_abs, relative=False, axis=_axis)
            else:
                container_obj = add_error_to_container("simple", container_obj, err_val=0.01 * _rel, relative=True)
                container_obj = add_error_to_container("simple", container_obj, err_val=_abs, relative=False)
            continue
        elif isinstance(_err, (float, int, str)):
            raise ValueError("Failed to read in errors: {}".format(_err))

        _add_kwargs = dict()
        # translate and check that all required keys are present
        try:
            _err_type = _err.get("type", "simple")

            _add_kwargs["name"] = _err.get("name", None)

            if _err_type == "simple":
                _add_kwargs["err_val"] = _err["error_value"]
                _add_kwargs["correlation"] = _err.get("correlation_coefficient", 0)
            elif _err_type == "matrix":
                _add_kwargs["err_matrix"] = _err["matrix"]
                _add_kwargs["matrix_type"] = _err["matrix_type"]
                # default None only mandatory for cor mats; check done later
                _add_kwargs["err_val"] = _err.get("error_value", None)
            else:
                raise TypeError("Unknown error type '{}'. " "Valid: {}".format(_err_type, ("simple", "matrix")))

            _add_kwargs["relative"] = _err.get("relative", False)

            # if needed, specify the axis (only for 'xy' containers)
            if _axis is not None:
                _add_kwargs["axis"] = _axis
        except KeyError as e:
            # KeyErrors mean the YAML is incomplete -> raise
            raise ValueError("Missing required key '%s' for error specification" % e.args[0])

        # add error to data container
        container_obj = add_error_to_container(_err_type, container_obj, **_add_kwargs)

    return container_obj, yaml_doc


class MatrixYamlDumper(yaml.Dumper):
    """Custom directives for writing out matrices"""

    _regex_space_after_newline = re.compile(r"\n\s+")  # match all spaces immediately after newline
    # match exactly 'n' sequences of non-whitespace followed by optional whitespace
    _regex_format_n_numbers = r"((\S+\s*){{}})"

    def matrix(self, numpy_matrix):
        """Represent a matrix as a space/endline separated string.
        Note: symmetric matrices are represented as a lower triangular matrix by default
        """
        _is_symmetric = np.allclose(numpy_matrix, numpy_matrix.T)

        # remove brackets
        _string_repr = str(numpy_matrix).replace("[[", " [").replace("]]", "] ").replace("[", "").replace("]", "").strip()
        # remove all spaces immediately after newline
        _string_repr = re.sub(self.__class__._regex_space_after_newline, "\n", _string_repr)

        # if symmetric, remove everything above upper diagonal
        if _is_symmetric and False:
            _rows = _string_repr.split("\n")
            for _irow, _row in enumerate(_rows):
                _rows[_irow] = re.sub(r"^((\S+\s*){{{}}}).*$".format(_irow + 1), r"\1", _row).strip()
            _string_repr = "\n".join(_rows)
            # write lower triangular matrix using the '|'-style
            return self.represent_scalar("!symmetric_matrix", _string_repr, style="|")

        # write full matrix using the '|'-style
        return self.represent_scalar("!matrix", _string_repr, style="|")


# representers for covariance matrices errors
MatrixYamlDumper.add_representer(np.array, MatrixYamlDumper.matrix)


class MatrixYamlLoader(yaml.Loader):
    """custom directives for reading in matrices"""

    def matrix(self, node):
        """construct a matrix from a tab/endline separated string"""
        _rows = node.value.split("\n")
        _mat = [_row.split() for _row in _rows if _row]  # last row may be empty -> leave out
        return np.array(_mat, dtype=float)

    def symmetric_matrix(self, node):
        """construct a lower triangular matrix from a tab/endline separated string"""
        _rows = node.value.split("\n")[:-1]  # last endline ends the matrix -> no new row
        _mat = [_row.split() for _row in _rows]
        _np_mat = np.array(np.zeros((len(_rows), len(_rows))), dtype=float)

        for _i, _row in enumerate(_mat):
            # check shape -> row index must match row length
            if len(_row) != _i + 1:
                raise ValueError(
                    "Cannot parse lower triangular matrix: " "row #{} should have length {}, got {} " "instead!".format(_i + 1, _i + 1, len(_row))
                )
            # fill matrix
            _np_mat[_i, 0 : len(_row)] = _row  # fill below diagonal

        _np_mat = _np_mat + np.tril(_np_mat, -1).T  # symmetrize

        return _np_mat


MatrixYamlLoader.add_constructor("!matrix", MatrixYamlLoader.matrix)
MatrixYamlLoader.add_constructor("!symmetric_matrix", MatrixYamlLoader.symmetric_matrix)



from __future__ import print_function

import numpy as np
from scipy.special import gammaincc, gammainccinv

__all__ = ["ConfidenceLevel"]


class ConfidenceLevel(object):
    """
    Helper class for handling the conversion from confidence levels to sigma values and vice versa.
    """

    def __init__(self, n_dimensions=1, cl=None, sigma=None, delta_nll=None):
        self.ndim = n_dimensions
        _num_spec_not_none = 0
        if cl is not None:
            _num_spec_not_none += 1
        if sigma is not None:
            _num_spec_not_none += 1
        if delta_nll is not None:
            _num_spec_not_none += 1
        if _num_spec_not_none != 1:
            raise ValueError("Exactly one out of cl, sigma, and delta_nll must be defined!")
        if cl is not None:
            self.cl = cl
        if sigma is not None:
            self.sigma = sigma
        if delta_nll is not None:
            self.sigma = np.sqrt(delta_nll)

    def __str__(self):
        return "<ConfidenceLevel (d=%d): %.4g%% (%.3g-sigma)>" % (
            self.ndim,
            self.cl * 100.0,
            self.sigma,
        )

    def _calc_sigma_from_cl(self):
        self._sigma = np.sqrt(2 * gammainccinv(self.ndim / 2.0, 1.0 - self.cl))

    def _calc_cl_from_sigma(self):
        self._cl = 1.0 - gammaincc(self.ndim / 2.0, self.sigma**2 / 2.0)

    @property
    def cl(self) -> float:
        if self._cl is None:
            self._calc_cl_from_sigma()
        return self._cl

    @cl.setter
    def cl(self, new_cl: float):
        if new_cl <= 0 or new_cl >= 1:
            raise ValueError("Confidence level must be greater than 0 and less than 1. Got: %g" % (new_cl,))
        self._cl = float(new_cl)
        self._sigma = None

    @property
    def sigma(self) -> float:
        if self._sigma is None:
            self._calc_sigma_from_cl()
        return self._sigma

    @sigma.setter
    def sigma(self, new_sigma: float):
        if new_sigma <= 0:
            raise ValueError("Sigma value must be greater than 0! Got: %g" % (new_sigma,))
        self._sigma = float(new_sigma)
        self._cl = None

    @property
    def delta_nll(self) -> float:
        return self.sigma**2

    @delta_nll.setter
    def delta_nll(self, new_delta_nll: float):
        if new_delta_nll <= 0:
            raise ValueError("delta_nll value must be greater than 0! Got: %g" % (new_delta_nll,))
        self.sigma = np.sqrt(new_delta_nll)

    @property
    def ndim(self):
        return self._ndim

    @ndim.setter
    def ndim(self, new_ndim):
        if not isinstance(new_ndim, int):
            raise ValueError("Number of dimensions must be of type int! Received type: %s" % type(new_ndim))
        if new_ndim <= 0:
            raise ValueError("Number of dimensions must be greater 0! Received: %d" % (new_ndim,))
        self._ndim = new_ndim

    @property
    def sigma_string(self):
        return "%.3g-sigma" % (self.sigma,)

    @property
    def sigma_latex_string(self):
        return r"%g$\sigma$" % (self.sigma,)

    @property
    def cl_string(self):
        return "%.4g%% CL" % (self.cl * 100,)

    @property
    def cl_latex_string(self):
        return r"$%.4g\%%$ CL" % (self.cl * 100,)



"""
The easiest way to use *kafe2* (as part of a *Python* program) is to use the wrapper functions
below.
These functions provide pre-configured pipelines for the most common use cases and do not require
the user to manually manage objects.
"""

__all__ = ["custom_fit", "hist_fit", "indexed_fit", "unbinned_fit", "xy_fit", "plot", "k2Fit"]

try:
    import typing  # help IDEs with type-hinting inside docstrings  # noqa: F401 (unused import)
except ImportError:
    pass

import os
import warnings
from copy import deepcopy
from glob import glob

import numpy as np

_fit_history = []


def _get_file_index():
    os.makedirs("results", exist_ok=True)
    _file_index = 0
    _globbed_files = glob(f"results/fit-{_file_index:04d}-*")
    while len(_globbed_files) > 0:
        _file_index += 1
        _globbed_files = glob(f"results/fit-{_file_index:04d}-*")
    return _file_index


def _fit_wrapper_generic(fit, p0, dp0, limits, fixed, constraints, report, profile, save):
    if p0 is not None:
        fit.set_all_parameter_values(p0)
    if dp0 is not None:
        fit.parameter_errors = dp0

    if limits is not None:
        if not isinstance(limits[0], (list, tuple)):
            limits = (limits,)
        for _limit in limits:
            fit.limit_parameter(*_limit)
    if fixed is not None:
        if not isinstance(fixed[0], (list, tuple)):
            fixed = (fixed,)
        for _fix in fixed:
            fit.fix_parameter(*_fix)
    if constraints is not None:
        if not isinstance(constraints[0], (list, tuple)):
            constraints = (constraints,)
        for _constraint in constraints:
            fit.add_parameter_constraint(*_constraint)

    _fit_results = fit.do_fit(asymmetric_parameter_errors=profile)
    _fit_results["fit"] = fit
    if report:
        fit.report(asymmetric_parameter_errors=profile)

    if save:
        _file_index = _get_file_index()
        fit.save_state(f"results/fit-{_file_index:04d}-results.yml")
        with open(f"results/fit-{_file_index:04d}-report.txt", "w", encoding="utf8") as _f:
            fit.report(_f, asymmetric_parameter_errors=profile)
    else:
        _file_index = None

    _fit_history.append(dict(fit=fit, profile=profile, file_index=_file_index))

    return _fit_results


def _add_error_to_fit_generic(fit, error, errors_rel_to_model, correlated=False, relative=False):
    if error is None:
        return
    error = np.asarray(error)
    _reference = "model" if errors_rel_to_model and relative else "data"
    if correlated:
        if error.ndim == 0:
            error = np.reshape(error, (1,))
        for _err in error:
            fit.add_error(_err, correlation=1.0, relative=relative, reference=_reference)
    else:
        if error.ndim == 2:
            fit.add_matrix_error(error, "cov", relative=relative, reference=_reference)
        else:
            fit.add_error(error, relative=relative, reference=_reference)


def custom_fit(
    cost_function,
    p0=None,
    dp0=None,
    limits=None,
    fixed=None,
    constraints=None,
    report=False,
    profile=True,
    save=True,
):
    """
    Built-in function for directly minimizing a cost function without any explicit model, data, or
    errors.

    :param cost_function: The cost function to be minimized as a native Python function.
    :type cost_function: typing.Callable
    :param p0: the initial parameter values for the fit.
    :type p0: typing.Sequence[float]
    :param dp0: the initial parameter step size for the fit.
    :type dp0: typing.Sequence[float]
    :param limits: limits to be applied to the model parameter. The expected format for each limit
        is an iterable consisting of the parameter name, the lower bound, and then the upper bound.
        An iterable of limits can be passed to limit multiple parameters.
    :type limits: typing.Sequence or typing.Sequence[typing.Union[list, tuple]]
    :param fixed: Model parameter to be fixed. The expected format for each parameter is the
        parameter name followed by an optional value to which the parameter should be set prior to
        fixing. An iterable of (name, value) tuples can be passed to fix multiple parameters.
    :type fixed: typing.Sequence or typing.Sequence[typing.Union[list, tuple]]
    :param constraints: constraints to be applied to the model parameter. The expected format for
        each constraint is an iterable consisting of the parameter name, the parameter mean, and
        then the parameter uncertainty. An iterable of constraints can be passed to limit multiple
        parameters.
    :type constraints: typing.Sequence or typing.Sequence[typing.Union[list, tuple]]
    :param report: whether a report of the data and fit results should be printed to the console.
    :type report: bool
    :param profile: whether the profile likelihood method should be used for asymmetric parameter
        errors and profile/contour plots.
    :type profile: bool
    :param save: whether the fit results should be saved to disk under `results`.
    :type save: bool
    :return: the fit results.
    :rtype: dict
    """
    from kafe2.fit.custom.fit import CustomFit

    _fit = CustomFit(cost_function)
    return _fit_wrapper_generic(_fit, p0, dp0, limits, fixed, constraints, report, profile, save)


def hist_fit(
    model_function=None,
    data=None,
    n_bins=None,
    bin_range=None,
    bin_edges=None,
    p0=None,
    dp0=None,
    error=None,
    error_rel=None,
    error_cor=None,
    error_cor_rel=None,
    errors_rel_to_model=True,
    density=True,
    gauss_approximation=None,
    limits=None,
    fixed=None,
    constraints=None,
    report=False,
    profile=True,
    save=True,
):
    """
    Built-in function for fitting a (probability density) function to one-dimensional data by
    binning the data. The uncertainty on the bins is assumed to follow a Poisson distribution. If
    any errors are specified then the Poisson distribution is instead approximated by a Gaussian
    distribution.

    :param model_function: The model function as a native Python function where the first
        argument denotes the independent *x* variable. Alternatively an already defined
        :py:class:`~kafe2.fit.histogram.model.HistModelFunction` object. Defaults to a normal
        distribution.
    :type model_function: typing.Callable
    :param data: the data for the fit. Can be either raw data, the result of `np.histogram`, or a
        :py:class:`kafe2.fit.hist.container.HistContainer` object.
    :type data: typing.Sequence[float] or :py:class:`kafe2.fit.hist.container.HistContainer`
    :param n_bins: how many bins raw data should be split into.
    :type n_bins: int
    :param bin_range: the lower and upper bound for the bins specified by n_bins.
    :type bin_range: typing.Sequence[float] of length 2
    :param bin_edges: explicit bin edges for raw data. If ``None``, each bin will have the same
        width.
    :type bin_edges: typing.Sequence[float]
    :param p0: the initial parameter values for the fit.
    :type p0: typing.Sequence[float]
    :param dp0: the initial parameter step size for the fit.
    :type dp0: typing.Sequence[float]
    :param error: uncorrelated absolute error on the bin heights.
    :type error: float or typing.Sequence[float]
    :param error_rel: uncorrelated relative error on the bin heights.
    :type error_rel: float or typing.Sequence[float]
    :param error_cor: correlated absolute error on the bin heights.
    :type error_cor: float or typing.Sequence[float]
    :param error_cor_rel: correlated relative error on the bin heights.
    :type error_cor_rel: float or typing.Sequence[float]
    :param errors_rel_to_model: whether the relative *y* errors should be relative to the model.
        Otherwise they are relative to the data.
    :type errors_rel_to_model: bool
    :param density: whether the model is a probability density function and the data should be
        normalized to match it.
    :type density: bool
    :param limits: limits to be applied to the model parameter. The expected format for each limit
        is an iterable consisting of the parameter name, the lower bound, and then the upper bound.
        An iterable of limits can be passed to limit multiple parameters.
    :type limits: typing.Sequence or typing.Sequence[typing.Union[list, tuple]]
    :param fixed: Model parameter to be fixed. The expected format for each parameter is the
        parameter name followed by an optional value to which the parameter should be set prior to
        fixing. An iterable of (name, value) tuples can be passed to fix multiple parameters.
    :type fixed: typing.Sequence or typing.Sequence[typing.Union[list, tuple]]
    :param constraints: constraints to be applied to the model parameter. The expected format for
        each constraint is an iterable consisting of the parameter name, the parameter mean, and
        then the parameter uncertainty. An iterable of constraints can be passed to limit multiple
        parameters.
    :type constraints: typing.Sequence or typing.Sequence[typing.Union[list, tuple]]
    :param report: whether a report of the data and fit results should be printed to the console.
    :type report: bool
    :param profile: whether the profile likelihood method should be used for asymmetric parameter
        errors and profile/contour plots.
    :type profile: bool
    :param save: whether the fit results should be saved to disk under `results`.
    :type save: bool
    :return: the fit results.
    :rtype: dict
    """
    from kafe2.fit.histogram.container import HistContainer
    from kafe2.fit.histogram.fit import HistFit

    data = HistContainer(n_bins, bin_range, bin_edges, data)

    if gauss_approximation is None:
        gauss_approximation = error is not None or error_rel is not None or error_cor is not None or error_cor_rel is not None
    _cost_function = "gauss_approximation" if gauss_approximation else "poisson"

    if model_function is None:
        _fit = HistFit(data, cost_function=_cost_function, density=density)
    else:
        _fit = HistFit(data, model_function, cost_function=_cost_function, density=density)

    _add_error_to_fit_generic(_fit, error, errors_rel_to_model)
    _add_error_to_fit_generic(_fit, error_cor, errors_rel_to_model, correlated=True)
    _add_error_to_fit_generic(_fit, error_rel, errors_rel_to_model, relative=True)
    _add_error_to_fit_generic(_fit, error_cor_rel, errors_rel_to_model, correlated=True, relative=True)

    return _fit_wrapper_generic(_fit, p0, dp0, limits, fixed, constraints, report, profile, save)


def indexed_fit(
    model_function=None,
    data=None,
    p0=None,
    dp0=None,
    error=None,
    error_rel=None,
    error_cor=None,
    error_cor_rel=None,
    errors_rel_to_model=True,
    limits=None,
    fixed=None,
    constraints=None,
    report=False,
    profile=True,
    save=True,
):
    from kafe2.fit.indexed import IndexedFit

    _fit = IndexedFit(data, model_function)

    _add_error_to_fit_generic(_fit, error, errors_rel_to_model)
    _add_error_to_fit_generic(_fit, error_cor, errors_rel_to_model, correlated=True)
    _add_error_to_fit_generic(_fit, error_rel, errors_rel_to_model, relative=True)
    _add_error_to_fit_generic(_fit, error_cor_rel, errors_rel_to_model, correlated=True, relative=True)

    return _fit_wrapper_generic(_fit, p0, dp0, limits, fixed, constraints, report, profile, save)


def unbinned_fit(
    model_function=None,
    data=None,
    p0=None,
    dp0=None,
    limits=None,
    fixed=None,
    constraints=None,
    report=False,
    profile=True,
    save=True,
):
    """
    Built-in function for directly fitting a probability density function to one-dimensional data
    without binning the data.

    :param model_function: The model function as a native Python function where the first
        argument denotes the independent *x* variable. Alternatively an already defined
        :py:class:`~kafe2.fit._base.model.ModelFunctionBase` object. Defaults to a straight line.
    :type model_function: typing.Callable
    :param data: the data values for the fit. Must be one-dimensional.
    :type data: typing.Sequence[float]
    :param p0: the initial parameter values for the fit.
    :type p0: typing.Sequence[float]
    :param dp0: the initial parameter step size for the fit.
    :type dp0: typing.Sequence[float]
    :param limits: limits to be applied to the model parameter. The expected format for each limit
        is an iterable consisting of the parameter name, the lower bound, and then the upper bound.
        An iterable of limits can be passed to limit multiple parameters.
    :type limits: typing.Sequence or typing.Sequence[typing.Union[list, tuple]]
    :param fixed: Model parameter to be fixed. The expected format for each parameter is the
        parameter name followed by an optional value to which the parameter should be set prior to
        fixing. An iterable of (name, value) tuples can be passed to fix multiple parameters.
    :type fixed: typing.Sequence or typing.Sequence[typing.Union[list, tuple]]
    :param constraints: constraints to be applied to the model parameter. The expected format for
        each constraint is an iterable consisting of the parameter name, the parameter mean, and
        then the parameter uncertainty. An iterable of constraints can be passed to limit multiple
        parameters.
    :type constraints: typing.Sequence or typing.Sequence[typing.Union[list, tuple]]
    :param report: whether a report of the data and fit results should be printed to the console.
    :type report: bool
    :param profile: whether the profile likelihood method should be used for asymmetric parameter
        errors and profile/contour plots.
    :type profile: bool
    :param save: whether the fit results should be saved to disk under `results`.
    :type save: bool
    :return: the fit results.
    :rtype: dict
    """
    from kafe2.fit.unbinned import UnbinnedFit

    _fit = UnbinnedFit(data, model_function)

    return _fit_wrapper_generic(_fit, p0, dp0, limits, fixed, constraints, report, profile, save)


def xy_fit(
    model_function=None,
    x_data=None,
    y_data=None,
    p0=None,
    dp0=None,
    x_error=None,
    y_error=None,
    x_error_rel=None,
    y_error_rel=None,
    x_error_cor=None,
    y_error_cor=None,
    x_error_cor_rel=None,
    y_error_cor_rel=None,
    errors_rel_to_model=True,
    limits=None,
    fixed=None,
    constraints=None,
    report=False,
    profile=None,
    save=True,
):
    """
    Built-in function for fitting a model function to xy data.

    Interpretation of x_error, y_error, x_error_rel, and y_error_rel:
    If the input error is a simple float it is broadcast across the entire data vector.
    If the input error is a one-dimensional vector it is interpreted as a pointwise error vector.
    If the input error is a two-dimensional matrix it is interpreted as a covariance matrix.

    Interpretation of x_error_cor, y_error_cor, x_error_cor_rel, and y_error_cor_rel:
    If the input error is a simple float it is broadcast across the entire data vector.
    If the input error is a one-dimensional vector then each individual value is added as
    a separate error that is being broadcast across the entire data vector.

    :param model_function: The model function as a native Python function where the first
        argument denotes the independent *x* variable. Alternatively an already defined
        :py:class:`~kafe2.fit._base.model.ModelFunctionBase` object. Defaults to a straight line.
    :type model_function: typing.Callable
    :param x_data: the x data values for the fit. Must be one-dimensional.
    :type x_data: typing.Sequence[float]
    :param y_data: the y data values for the fit. Must be one-dimensional.
    :type y_data: typing.Sequence[float]
    :param p0: the initial parameter values for the fit.
    :type p0: typing.Sequence[float]
    :param dp0: the initial parameter step size for the fit.
    :type dp0: typing.Sequence[float]
    :param x_error: uncorrelated absolute *x* error.
    :type x_error: float or typing.Sequence[float]
    :param y_error: uncorrelated absolute *y* error.
    :type y_error: float or typing.Sequence[float]
    :param x_error_rel: uncorrelated relative *x* error.
    :type x_error_rel: float or typing.Sequence[float]
    :param y_error_rel: uncorrelated relative *y* error.
    :type y_error_rel: float or typing.Sequence[float]
    :param x_error_cor: correlated absolute *x* error.
    :type x_error_cor: float or typing.Sequence[float]
    :param y_error_cor: correlated absolute *y* error.
    :type y_error_cor: float or typing.Sequence[float]
    :param x_error_cor_rel: correlated relative *x* error.
    :type x_error_cor_rel: float or typing.Sequence[float]
    :param y_error_cor_rel: correlated relative *y* error.
    :type y_error_cor_rel: float or typing.Sequence[float]
    :param errors_rel_to_model: whether the relative *y* errors should be relative to the model.
        Otherwise they are relative to the data.
    :type errors_rel_to_model: bool
    :param limits: limits to be applied to the model parameter. The expected format for each limit
        is an iterable consisting of the parameter name, the lower bound, and then the upper bound.
        An iterable of limits can be passed to limit multiple parameters.
    :type limits: typing.Sequence or typing.Sequence[typing.Union[list, tuple]]
    :param fixed: Model parameter to be fixed. The expected format for each parameter is the
        parameter name followed by an optional value to which the parameter should be set prior to
        fixing. An iterable of (name, value) tuples can be passed to fix multiple parameters.
    :type fixed: typing.Sequence or typing.Sequence[typing.Union[list, tuple]]
    :param constraints: constraints to be applied to the model parameter. The expected format for
        each constraint is an iterable consisting of the parameter name, the parameter mean, and
        then the parameter uncertainty. An iterable of constraints can be passed to limit multiple
        parameters.
    :type constraints: typing.Sequence or typing.Sequence[typing.Union[list, tuple]]
    :param report: whether a report of the data and fit results should be printed to the console.
    :type report: bool
    :param profile: whether the profile likelihood method should be used for asymmetric parameter
        errors and profile/contour plots.
    :type profile: bool
    :param save: whether the fit results should be saved to disk under `results`.
    :type save: bool
    :return: the fit results.
    :rtype: dict
    """
    from kafe2.fit.xy.fit import XYFit

    if model_function is None:
        _fit = XYFit([x_data, y_data])
    else:
        _fit = XYFit([x_data, y_data], model_function)

    def _add_error_to_fit(axis, error, correlated=False, relative=False):
        if error is None:
            return
        error = np.asarray(error)
        _reference = "model" if errors_rel_to_model and axis == "y" and relative else "data"
        if correlated:
            if error.ndim == 0:
                error = np.reshape(error, (1,))
            for _err in error:
                _fit.add_error(axis, _err, correlation=1.0, relative=relative, reference=_reference)
        else:
            if error.ndim == 2:
                _fit.add_matrix_error(axis, error, "cov", relative=relative, reference=_reference)
            else:
                _fit.add_error(axis, error, relative=relative, reference=_reference)

    _add_error_to_fit("x", x_error)
    _add_error_to_fit("y", y_error)
    _add_error_to_fit("x", x_error_rel, relative=True)
    _add_error_to_fit("y", y_error_rel, relative=True)
    _add_error_to_fit("x", x_error_cor, correlated=True)
    _add_error_to_fit("y", y_error_cor, correlated=True)
    _add_error_to_fit("x", x_error_cor_rel, correlated=True, relative=True)
    _add_error_to_fit("y", y_error_cor_rel, correlated=True, relative=True)

    if profile is None:
        profile = x_error is not None or x_error_rel is not None or y_error_rel is not None

    return _fit_wrapper_generic(_fit, p0, dp0, limits, fixed, constraints, report, profile, save)


def plot(
    fits=-1,
    x_label=None,
    y_label=None,
    data_label=None,
    model_label=None,
    error_band_label=None,
    x_range=None,
    y_range=None,
    x_scale=None,
    y_scale=None,
    x_ticks=None,
    y_ticks=None,
    parameter_names=None,
    model_name=None,
    model_expression=None,
    font_scale=1.0,
    legend=True,
    fit_info=True,
    error_band=True,
    extra=None,
    profile=None,
    plot_profile=None,
    show=True,
    save=True,
):
    """
    Plots kafe2 fits.

    :param fits: which kafe2 fits to use for the plot. A positive integer is interpreted as the fit
        with the given index that has been performed (with wrappers) since the program started. A
        negative integer *-n* is interpreted as the last *n* fits. kafe2 fit objects are used
        directly.
    :type fits: int or :py:class:`~kafe2.fit._base.FitBase`
        or Sequence[:py:class:`~kafe2.fit._base.FitBase`]
    :param x_label: the *x* axis label.
    :type x_label: str
    :param y_label: the *y* axis label.
    :type y_label: str
    :param data_label: the data label(s) in the legend.
    :type data_label: str or typing.Sequence[str]
    :param model_label: the model label(s) in the legend (under data label).
    :type model_label: str or typing.Sequence[str]
    :param error_band_label: the error band label(s) in the legend.
    :type error_band_label: str or typing.Sequence[str]
    :param x_range: *x* range for the plot.
    :type x_range: typing.Sequence[float], len(x_range) == 2
    :param y_range: *y* range for the plot.
    :type y_range: typing.Sequence[float], len(y_range) == 2
    :param x_scale: the scale to use for the *x* axis.
    :type x_scale: "linear" or "log"
    :param y_scale: the scale to use for the *y* axis.
    :type y_scale: "linear" or "log"
    :param x_ticks: the ticks at which to show values on the *x* axis.
    :type x_ticks: typing.Sequence[float]
    :param y_ticks: the ticks at which to show values on the *y* axis.
    :type y_ticks: typing.Sequence[float]
    :param parameter_names: custom parameter LaTeX names to display in the plot. The dictionary keys
        are the regular parameter names and the dictionary values are the names to show in the plot.
    :type parameter_names: dict
    :param model_name: the model LaTeX name(s) in the legend (in the mathematical expression of the
        model function).
    :type model_name: str or typing.Sequence[str]
    :param model_expression: the model LaTeX expression(s) in the legend.
    :type model_expression: str or typing.Sequence[str]
    :param legend: whether the legend should be shown.
    :type legend: bool
    :param fit_info: whether the fit information (fit results, goodness of fit) should be shown.
    :type fit_info: bool
    :param error_band: whether the model error band should be shown.
    :type error_band: bool
    :param extra: additional, supplementary plots to show below the main plot.
    :type error_band: "ratio", "residual", or "pull".
    :param profile: whether the profile likelihood method should be used for asymmetric parameter
        errors and profile/contour plots.
    :type profile: bool
    :param plot_profile: whether the profile plots should be created.
    :type plot_profile: bool
    :param show: whether the plots should be shown.
    :type show: bool
    :param save: whether the plots should be saved to disk under `results`.
    :type save: bool
    :param font_scale: multiply font size by this amount.
    :type font_scale: float

    :return: a *kafe2* plot object containing the relevant matplotlib plots.
    :rtype: :py:class:`~kafe2.fit._base.Plot`
    """
    from kafe2 import ContoursProfiler, CustomFit, Plot

    _start_index = _get_file_index()

    _fit_profiles = None
    _file_indices = None
    if isinstance(fits, int):
        if fits >= 0:
            _fit_profiles = [_fit_history[fits]["profile"]]
            _file_indices = [_fit_history[fits]["file_index"]]
            fits = [_fit_history[fits]["fit"]]
        else:
            fits = _fit_history[fits:]
            _fit_profiles = [_f["profile"] for _f in fits]
            _file_indices = [_f["file_index"] for _f in fits]
            fits = [_f["fit"] for _f in fits]
    else:
        try:
            iter(fits)
        except TypeError:
            fits = [fits]
    if profile is None:
        profile = np.any(_fit_profiles)

    if plot_profile is None and _fit_profiles is not None:
        plot_profile = _fit_profiles
    if profile and plot_profile is not None:
        try:
            iter(plot_profile)
        except TypeError:
            plot_profile = [plot_profile for _ in fits]
        for _i, (_f_i, _pp_i) in enumerate(zip(fits, plot_profile)):
            if not _pp_i:
                continue
            _cpf = ContoursProfiler(_f_i)
            _cpf.plot_profiles_contours_matrix(font_scale=font_scale)
            if save:
                _file_index = _start_index + _i if _file_indices is None else _file_indices[_i]
                _cpf.save(f"results/fit-{_file_index:04d}-profile.png", dpi=240)

    fits = [_f for _f in fits if not isinstance(_f, CustomFit)]  # No sensible way to plot this.
    if parameter_names is not None:
        _unused_parameter_names = deepcopy(parameter_names)
        for _f in fits:
            _plns = {_p: _pn for _p, _pn in parameter_names.items() if _p in _f.parameter_names}
            _f.assign_parameter_latex_names(**_plns)
            for _parameter_name in _f.parameter_names:
                _unused_parameter_names.pop(_parameter_name, None)
            for _x_name in _f._model_function.x_name:
                _unused_parameter_names.pop(_x_name, None)
        if _unused_parameter_names:
            warnings.warn(f"Unused parameter names for plot: {_unused_parameter_names}")
    if model_name is not None:
        if isinstance(model_name, str):
            model_name = [model_name for _ in fits]
        for _f, _mn in zip(fits, model_name):
            _f.assign_model_function_latex_name(_mn)
    if model_expression is not None:
        if isinstance(model_expression, str):
            model_expression = [model_expression for _ in fits]
        for _f, _me in zip(fits, model_expression):
            _f.assign_model_function_latex_expression(_me)
    _plot = Plot(fits)
    if x_label is not None:
        _plot.x_label = x_label
    if y_label is not None:
        _plot.y_label = y_label
    if data_label is not None:
        _plot.customize("data", "label", data_label)
    if model_label is not None:
        _plot.customize("model", "label", model_label)
        # XYFit for some reason uses a different keyword:
        try:
            _plot.customize("model_line", "label", model_label)
        except ValueError:
            pass
        # Apply model label to density label in plot if HistFit:
        try:
            _plot.customize("model_density", "label", f"{model_label} density")
        except ValueError:
            pass
    if error_band_label is not None:
        _plot.customize("model_error_band", "label", error_band_label)
    if model_label is not None and error_band_label is None:
        error_band_label = f"{model_label} $\\pm 1 \\sigma$"
        try:
            _plot.customize("model_error_band", "label", error_band_label)
        except ValueError:
            pass
    if not error_band:
        _plot.customize("model_error_band", "label", None)
        _plot.customize("model_error_band", "hide", True)

    if x_range is not None:
        _plot.x_range = x_range
    if y_range is not None:
        _plot.y_range = y_range
    if x_scale is not None:
        _plot.x_scale = x_scale
    if y_scale is not None:
        _plot.y_scale = y_scale
    if x_ticks is not None:
        _plot.x_ticks = x_ticks
    if y_ticks is not None:
        _plot.y_ticks = y_ticks

    if extra not in [None, "ratio", "residual", "pull"]:
        raise ValueError(f"Unknown extra plot: '{extra}'. Available: None, 'ratio', 'residual', 'pull'.")
    if len(fits) > 0:  # Do not plot if only CustomFit.
        _plot.plot(
            legend=legend,
            fit_info=fit_info,
            asymmetric_parameter_errors=profile,
            font_scale=font_scale,
            ratio=extra == "ratio",
            residual=extra == "residual",
            pull=extra == "pull",
        )

        if save:
            for _i, _ in enumerate(fits):
                _file_index = _start_index + _i if _file_indices is None else _file_indices[_i]
                _plot.save(f"results/fit-{_file_index:04d}-plot.png", dpi=240)

    if show:
        _plot.show()

    return _plot


_plot_func = plot


# def plot_xy_data(x_data, y_data, x_error=None, y_error=None, x_error_rel=None, y_error_rel=None,
#                 x_label=None, y_label=None, data_label=None, x_range=None, y_range=None,
#                 x_scale=None, y_scale=None, x_ticks=None, y_ticks=None, show=True, save=True):
#    from kafe2.fit.xy.container import XYContainer
#    _container = XYContainer(x_data, y_data)
#    if x_error is not None:
#        _container.add_error("x", x_error)
#    if y_error is not None:
#        _container.add_error("y", y_error)
#    if x_error_rel is not None:
#        _container.add_error("x", x_error_rel, relative=True)
#    if y_error_rel is not None:
#        _container.add_error("y", y_error_rel, relative=True)
#    plot(_container, x_label=x_label, y_label=y_label, data_label=data_label, x_range=x_range,
#         y_range=y_range, x_scale=x_scale, y_scale=y_scale, x_ticks=x_ticks, y_ticks=y_ticks,
#         show=show, save=save)


def k2Fit(
    func,
    x,
    y,
    sx=None,
    sy=None,
    srelx=None,
    srely=None,
    xabscor=None,
    yabscor=None,
    xrelcor=None,
    yrelcor=None,
    ref_to_model=True,
    constraints=None,
    p0=None,
    dp0=None,
    limits=None,
    plot=True,
    axis_labels=["x-data", "y-data"],
    data_legend="data",
    model_expression=None,
    model_name=None,
    model_legend="model",
    model_band=r"$\pm 1 \sigma$",
    fit_info=True,
    plot_band=True,
    asym_parerrs=True,
    plot_cor=False,
    showplots=True,
    quiet=True,
):
    """
    Legacy function for backwards compatibility with *PhyPraKit*.
    **New code should not use this function.**
    Fits a model to *xy* data and plots the results.

    Interpretation of sx, sy, srelx, and srely:
    If the input error is a simple float it is broadcast across the entire data vector.
    If the input error is a one-dimensional vector it is interpreted as a pointwise error vector.
    If the input error is a two-dimensional matrix it is interpreted as a covariance matrix.

    Interpretation of xabscor, yabscor, xrelcor, and yrelcor:
    If the input error is a simple float it is broadcast across the entire data vector.
    If the input error is a one-dimensional vector then each individual value is added as
    a separate error that is being broadcast across the entire data vector.

    :param func: The model function as a native Python function where the first
        argument denotes the independent *x* variable. Alternatively an already defined
        :py:class:`~kafe2.fit._base.model.ModelFunctionBase` object. Defaults to a straight line.
    :type func: typing.Callable
    :param x: the *x* data values for the fit. Must be one-dimensional.
    :type x: typing.Sequence[float]
    :param y: the *y* data values for the fit. Must be one-dimensional.
    :type y: typing.Sequence[float]
    :param sx: uncorrelated absolute *x* error.
    :type sx: float or typing.Sequence[float]
    :param sy: uncorrelated absolute *y* error.
    :type sy: float or typing.Sequence[float]
    :param srelx: uncorrelated relative *x* error.
    :type srelx: float or typing.Sequence[float]
    :param srely: uncorrelated relative *y* error.
    :type srely: float or typing.Sequence[float]
    :param xabscor: correlated absolute *x* error.
    :type xabscor: float or typing.Sequence[float]
    :param yabscor: correlated absolute *y* error.
    :type yabscor: float or typing.Sequence[float]
    :param xrelcor: correlated relative *x* error.
    :type xrelcor: float or typing.Sequence[float]
    :param yrelcor: correlated relative *y* error.
    :type yrelcor: float or typing.Sequence[float]
    :param ref_to_model: whether the relative *y* errors should be relative to the model.
        Otherwise they are relative to the data.
    :type ref_to_model: bool
    :param constraints: constraints to be applied to the model parameter. The expected format for
        each constraint is an iterable consisting of the parameter name, the parameter mean, and
        then the parameter uncertainty. An iterable of constraints can be passed to limit multiple
        parameters.
    :type constraints: typing.Sequence or typing.Sequence[typing.Union[list, tuple]]
    :param p0: the initial parameter values for the fit.
    :type p0: typing.Sequence[float]
    :param dp0: the initial parameter step size for the fit.
    :type dp0: typing.Sequence[float]
    :param limits: limits to be applied to the model parameter. The expected format for each limit
        is an iterable consisting of the parameter name, the lower bound, and then the upper bound.
        An iterable of limits can be passed to limit multiple parameters.
    :type limits: typing.Sequence or typing.Sequence[typing.Union[list, tuple]]
    :param plot: whether the fit results should be plotted.
    :type plot: bool
    :param axis_labels: the labels for the *x* and *y* axis.
    :type axis_labels: typing.Sequence[str]
    :param data_legend: the data label in the legend.
    :type data_legend: str
    :param model_expression: the model LaTeX expression in the legend.
    :type model_expression: str
    :param model_name: the model LaTeX name in the legend (in the mathematical expression of the
        model function).
    :type model_name: str
    :param model_legend: the model label in the legend (under data label).
    :type model_legend: str
    :param model_band: the error band label in the legend.
    :type model_band: str
    :param fit_info: whether the fit information (fit results, goodness of fit) should be shown.
    :type fit_info: bool
    :param plot_band: whether the model error band should be shown.
    :type plot_band: bool
    :param asym_parerrs: whether the profile likelihood method should be used for asymmetric
        parameter errors.
    :type asym_parerrs: bool
    :param plot_cor: whether the profile plots should be created.
    :type plot_cor: bool
    :param showplots: whether the plots should be shown.
    :type showplots: bool
    :param report: whether the report of the data and fit results should be suppressed.
    :type report: bool
    :return: a tuple containing the parameter values, the parameter errors, the parameter
        correlation matrix, and the minimal :math:`\\chi^2` cost function value.
    :rtype: tuple
    """
    xy_fit(
        func,
        x,
        y,
        p0=p0,
        dp0=dp0,
        x_error=sx,
        y_error=sy,
        x_error_rel=srelx,
        y_error_rel=srely,
        x_error_cor=xabscor,
        y_error_cor=yabscor,
        x_error_cor_rel=xrelcor,
        y_error_cor_rel=yrelcor,
        errors_rel_to_model=ref_to_model,
        limits=limits,
        constraints=constraints,
        report=not quiet,
        profile=True,
    )
    if plot:
        _plot_func(
            x_label=axis_labels[0],
            y_label=axis_labels[1],
            data_label=data_legend,
            model_label=model_legend,
            error_band_label=model_band,
            model_name=model_name,
            model_expression=model_expression,
            legend=True,
            fit_info=fit_info,
            error_band=plot_band,
            profile=True,
            plot_profile=plot_cor,
            show=showplots,
        )
    _fit_object = _fit_history[-1]["fit"]
    _parameter_errors = (
        _fit_object.asymmetric_parameter_errors if asym_parerrs else np.stack([-_fit_object.parameter_errors, _fit_object.parameter_errors], axis=-1)
    )
    return (
        _fit_object.parameter_values,
        _parameter_errors,
        _fit_object.parameter_cor_mat,
        _fit_object.goodness_of_fit,
    )



import abc

import six

# import data container classes
from ...histogram import HistContainer
from ...indexed import IndexedContainer
from ...unbinned import UnbinnedContainer
from ...xy import XYContainer
from .._base import GenericDReprBase

__all__ = ["DataContainerDReprBase"]


@six.add_metaclass(abc.ABCMeta)
class DataContainerDReprBase(GenericDReprBase):
    BASE_OBJECT_TYPE_NAME = "container"

    _CLASS_TO_OBJECT_TYPE_NAME = {
        HistContainer: "histogram",
        IndexedContainer: "indexed",
        UnbinnedContainer: "unbinned",
        XYContainer: "xy",
    }
    _OBJECT_TYPE_NAME_TO_CLASS = {
        "histogram": HistContainer,
        "indexed": IndexedContainer,
        "unbinned": UnbinnedContainer,
        "xy": XYContainer,
    }

    def __init__(self, data_container=None):
        self._kafe_object = data_container
        super(DataContainerDReprBase, self).__init__()



import unittest

import numpy as np
from six import StringIO

from kafe2.fit._base import ModelFunctionBase
from kafe2.fit.histogram import HistModelFunction
from kafe2.fit.histogram.model import HistParametricModel
from kafe2.fit.indexed import IndexedModelFunction
from kafe2.fit.indexed.model import IndexedParametricModel
from kafe2.fit.io.handle import IOStreamHandle
from kafe2.fit.representation import (
    ParametricModelYamlReader,
    ParametricModelYamlWriter,
)
from kafe2.fit.xy.model import XYParametricModel

TEST_PARAMETRIC_MODEL_HIST = """
type: histogram
n_bins: 5
bin_range: [0, 5]
model_density_function:
    type: histogram
    python_code: |
        def linear_model(x, a, b):
            return a * x + b
model_parameters: [0.0, 0.08]
"""

TEST_PARAMETRIC_MODEL_HIST_MISSING_KEYWORD = """
type: histogram
n_bins: 5
bin_range: [0, 5]
model_parameters: [0.0, 0.08]
"""

TEST_PARAMETRIC_MODEL_HIST_EXTRA_KEYWORD = (
    TEST_PARAMETRIC_MODEL_HIST
    + """
extra_keyword: 3.14
"""
)

TEST_PARAMETRIC_MODEL_HIST_WITH_ERRORS = (
    TEST_PARAMETRIC_MODEL_HIST
    + """
errors:
  - correlation_coefficient: 0.0
    error_value: 0.1
    name: test_error
    relative: false
    type: simple
"""
)

TEST_PARAMETRIC_MODEL_HIST_LEGACY = (
    TEST_PARAMETRIC_MODEL_HIST
    + """
model_density_func_antiderivative: null
"""
)

TEST_PARAMETRIC_MODEL_HIST_ANTIDERIVATIVE = (
    TEST_PARAMETRIC_MODEL_HIST
    + """
bin_evaluation: |
    def linear_model_antiderivative(x, a, b):
        return 0.5 * a * x ** 2 + b * x
"""
)

TEST_PARAMETRIC_MODEL_HIST_NUMPY_VECTORIZE = (
    TEST_PARAMETRIC_MODEL_HIST
    + """
bin_evaluation: |
    @np.vectorize
    def linear_model_antiderivative(x, a, b):
        return 0.5 * a * x ** 2 + b * x
"""
)


class TestHistParametricModelYamlRepresenter(unittest.TestCase):
    def _assert_read_model_correct(self, read_model, ref_model, errors=False):
        self.assertTrue(isinstance(read_model, HistParametricModel))

        self.assertTrue(np.all(read_model.data == ref_model.data))
        if isinstance(ref_model.bin_evaluation, str):
            self.assertEqual(read_model.bin_evaluation, ref_model.bin_evaluation)
            self.assertEqual(read_model.bin_evaluation_string, ref_model.bin_evaluation_string)
        self.assertTrue(np.all(read_model.parameters == self._test_model_parameters))
        self.assertTrue(read_model.n_bins == self._test_n_bins)
        self.assertTrue(np.all(read_model.bin_range == self._test_bin_range))
        self.assertTrue(np.all(read_model.bin_edges == self._test_bin_edges))
        self.assertTrue(
            np.allclose(
                read_model.eval_model_function_density(self._test_x),
                ref_model.eval_model_function_density(self._test_x),
            )
        )

        if errors:
            _given_error = self._test_parametric_model_with_errors.get_error("test_error")
            _read_error = read_model.get_error("test_error")
            self.assertTrue(_given_error["enabled"] == _read_error["enabled"])
            self.assertTrue(np.allclose(_given_error["err"].error, _read_error["err"].error))
            self.assertTrue(_given_error["err"].corr_coeff == _read_error["err"].corr_coeff)

    @staticmethod
    def _get_model_from_string(input_string):
        _testfile_stringstream = IOStreamHandle(StringIO(input_string))
        return ParametricModelYamlReader(_testfile_stringstream).read()

    @staticmethod
    def _get_model_from_roundtrip(model):
        _roundtrip_stringstream = IOStreamHandle(StringIO())
        _roundtrip_streamreader = ParametricModelYamlReader(_roundtrip_stringstream)
        _roundtrip_streamwriter = ParametricModelYamlWriter(model, _roundtrip_stringstream)
        _roundtrip_streamwriter.write()
        _roundtrip_stringstream.seek(0)  # return to beginning
        return _roundtrip_streamreader.read()

    @staticmethod
    def linear_model(x, a, b):
        return a * x + b

    @staticmethod
    def linear_model_antiderivative(x, a, b):
        return 0.5 * a * x**2 + b * x

    @staticmethod
    @np.vectorize
    def linear_model_numpy_vectorize(x, a, b):
        return 0.5 * a * x**2 + b * x

    def setUp(self):
        self._test_n_bins = 5
        self._test_bin_range = (0, 5)
        self._test_bin_edges = np.arange(6)
        self._test_model_parameters = np.array([0.0, 0.08])
        self._test_x = np.linspace(start=0.0, stop=5.0, num=11, endpoint=True)
        self._test_parametric_model = HistParametricModel(
            n_bins=self._test_n_bins,
            bin_range=self._test_bin_range,
            model_density_func=HistModelFunction(TestHistParametricModelYamlRepresenter.linear_model),
            model_parameters=self._test_model_parameters,
        )

        self._test_parametric_model_with_errors = HistParametricModel(
            n_bins=self._test_n_bins,
            bin_range=self._test_bin_range,
            model_density_func=HistModelFunction(TestHistParametricModelYamlRepresenter.linear_model),
            model_parameters=self._test_model_parameters,
        )
        self._test_parametric_model_with_errors.add_error(err_val=0.1, name="test_error", correlation=0, relative=False)

        self._test_parametric_model_antiderivative = HistParametricModel(
            n_bins=self._test_n_bins,
            bin_range=self._test_bin_range,
            model_density_func=HistModelFunction(TestHistParametricModelYamlRepresenter.linear_model),
            model_parameters=self._test_model_parameters,
            bin_evaluation=TestHistParametricModelYamlRepresenter.linear_model_antiderivative,
        )

        self._test_parametric_model_numpy_vectorize = HistParametricModel(
            n_bins=self._test_n_bins,
            bin_range=self._test_bin_range,
            model_density_func=HistModelFunction(TestHistParametricModelYamlRepresenter.linear_model),
            model_parameters=self._test_model_parameters,
            bin_evaluation=TestHistParametricModelYamlRepresenter.linear_model_numpy_vectorize,
        )

    def test_read_from_testfile_stream(self):
        _read_parametric_model = TestHistParametricModelYamlRepresenter._get_model_from_string(TEST_PARAMETRIC_MODEL_HIST)
        self._assert_read_model_correct(_read_parametric_model, self._test_parametric_model)

    def test_read_from_testfile_stream_antiderivative(self):
        _read_parametric_model = TestHistParametricModelYamlRepresenter._get_model_from_string(TEST_PARAMETRIC_MODEL_HIST_ANTIDERIVATIVE)
        self._assert_read_model_correct(_read_parametric_model, self._test_parametric_model_antiderivative)

    def test_read_from_testfile_stream_numpy_vectorize(self):
        _read_parametric_model = TestHistParametricModelYamlRepresenter._get_model_from_string(TEST_PARAMETRIC_MODEL_HIST_NUMPY_VECTORIZE)
        self._assert_read_model_correct(_read_parametric_model, self._test_parametric_model_numpy_vectorize)

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            _read_parametric_model = TestHistParametricModelYamlRepresenter._get_model_from_string(TEST_PARAMETRIC_MODEL_HIST_MISSING_KEYWORD)

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            _read_parametric_model = TestHistParametricModelYamlRepresenter._get_model_from_string(TEST_PARAMETRIC_MODEL_HIST_EXTRA_KEYWORD)

    def test_read_from_testfile_stream_with_errors(self):
        _read_parametric_model = TestHistParametricModelYamlRepresenter._get_model_from_string(TEST_PARAMETRIC_MODEL_HIST_WITH_ERRORS)
        self._assert_read_model_correct(_read_parametric_model, self._test_parametric_model_with_errors, errors=True)

    def test_read_from_testfile_stream_legacy(self):
        _read_parametric_model = TestHistParametricModelYamlRepresenter._get_model_from_string(TEST_PARAMETRIC_MODEL_HIST_LEGACY)
        self._assert_read_model_correct(_read_parametric_model, self._test_parametric_model)

    def test_round_trip_with_stringstream(self):
        _read_parametric_model = self._get_model_from_roundtrip(self._test_parametric_model)
        self._assert_read_model_correct(_read_parametric_model, self._test_parametric_model)

    def test_round_trip_with_stringstream_antiderivative(self):
        _read_parametric_model = self._get_model_from_roundtrip(self._test_parametric_model_antiderivative)
        self._assert_read_model_correct(_read_parametric_model, self._test_parametric_model_antiderivative)

    def test_round_trip_with_stringstream_numpy_vectorize(self):
        _read_parametric_model = self._get_model_from_roundtrip(self._test_parametric_model_numpy_vectorize)
        self._assert_read_model_correct(_read_parametric_model, self._test_parametric_model_numpy_vectorize)

    def test_round_trip_with_stringstream_with_errors(self):
        _read_parametric_model = self._get_model_from_roundtrip(self._test_parametric_model_with_errors)
        self._assert_read_model_correct(_read_parametric_model, self._test_parametric_model_with_errors, errors=True)


TEST_PARAMETRIC_MODEL_INDEXED = """
type: indexed
model_function:
    type: indexed
    python_code: |
        def linear_model(a, b):
            return a * np.arange(6) + b
model_parameters: [1.1, -1.5]
"""

TEST_PARAMETRIC_MODEL_INDEXED_MISSING_KEYWORD = """
type: indexed
model_parameters: [1.1, -1.5]
"""

TEST_PARAMETRIC_MODEL_INDEXED_EXTRA_KEYWORD = (
    TEST_PARAMETRIC_MODEL_INDEXED
    + """
extra_keyword: 3.14
"""
)

TEST_PARAMETRIC_MODEL_INDEXED_WITH_ERRORS = (
    TEST_PARAMETRIC_MODEL_INDEXED
    + """
errors:
  - correlation_coefficient: 0.0
    error_value: 0.1
    name: test_error
    relative: false
    type: simple
"""
)


class TestIndexedParametricModelYamlRepresenter(unittest.TestCase):
    TEST_X = np.arange(6)

    @staticmethod
    def linear_model(a, b):
        return a * np.arange(6) + b

    def setUp(self):
        self._test_model_parameters = np.array([1.1, -1.5])
        self._test_parametric_model = IndexedParametricModel(
            IndexedModelFunction(TestIndexedParametricModelYamlRepresenter.linear_model),
            self._test_model_parameters,
        )

        self._test_parametric_model_with_errors = IndexedParametricModel(
            IndexedModelFunction(TestIndexedParametricModelYamlRepresenter.linear_model),
            self._test_model_parameters,
        )
        self._test_parametric_model_with_errors.add_error(err_val=0.1, name="test_x_error", correlation=0, relative=False)

        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._roundtrip_stringstream_with_errors = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_INDEXED))
        self._testfile_stringstream_with_errors = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_INDEXED_WITH_ERRORS))

        self._roundtrip_streamreader = ParametricModelYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamreader_with_errors = ParametricModelYamlReader(self._roundtrip_stringstream_with_errors)
        self._roundtrip_streamwriter = ParametricModelYamlWriter(self._test_parametric_model, self._roundtrip_stringstream)
        self._roundtrip_streamwriter_with_errors = ParametricModelYamlWriter(
            self._test_parametric_model_with_errors, self._roundtrip_stringstream_with_errors
        )
        self._testfile_streamreader = ParametricModelYamlReader(self._testfile_stringstream)
        self._testfile_streamreader_with_errors = ParametricModelYamlReader(self._testfile_stringstream_with_errors)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_INDEXED_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_INDEXED_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ParametricModelYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ParametricModelYamlReader(self._testfile_stringstream_extra_keyword)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_write_to_roundtrip_stringstream_with_errors(self):
        self._roundtrip_streamwriter_with_errors.write()

    def test_read_from_testfile_stream(self):
        _read_parametric_model = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_parametric_model, IndexedParametricModel))

        self.assertTrue(np.allclose(_read_parametric_model.parameters, self._test_model_parameters))
        self.assertTrue(
            np.allclose(
                _read_parametric_model.eval_model_function(),
                self._test_parametric_model.eval_model_function(),
            )
        )

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_read_from_testfile_stream_with_errors(self):
        _read_parametric_model = self._testfile_streamreader_with_errors.read()
        self.assertTrue(isinstance(_read_parametric_model, IndexedParametricModel))

        self.assertTrue(np.allclose(_read_parametric_model.parameters, self._test_model_parameters))
        self.assertTrue(
            np.allclose(
                _read_parametric_model.eval_model_function(),
                self._test_parametric_model_with_errors.eval_model_function(),
            )
        )

        _given_error = self._test_parametric_model_with_errors.get_error("test_x_error")
        _read_error = _read_parametric_model.get_error("test_error")
        self.assertTrue(_given_error["enabled"] == _read_error["enabled"])
        self.assertTrue(np.allclose(_given_error["err"].error, _read_error["err"].error))
        self.assertTrue(_given_error["err"].corr_coeff == _read_error["err"].corr_coeff)

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_parametric_model = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_parametric_model, IndexedParametricModel))

        self.assertTrue(np.allclose(_read_parametric_model.parameters, self._test_model_parameters))
        self.assertTrue(
            np.allclose(
                _read_parametric_model.eval_model_function(),
                self._test_parametric_model.eval_model_function(),
            )
        )

    def test_round_trip_with_stringstream_with_errors(self):
        self._roundtrip_streamwriter_with_errors.write()
        self._roundtrip_stringstream_with_errors.seek(0)  # return to beginning
        _read_parametric_model = self._roundtrip_streamreader_with_errors.read()
        self.assertTrue(isinstance(_read_parametric_model, IndexedParametricModel))

        self.assertTrue(np.allclose(_read_parametric_model.parameters, self._test_model_parameters))
        self.assertTrue(
            np.allclose(
                _read_parametric_model.eval_model_function(),
                self._test_parametric_model_with_errors.eval_model_function(),
            )
        )

        _given_error = self._test_parametric_model_with_errors.get_error("test_x_error")
        _read_error = _read_parametric_model.get_error("test_x_error")
        self.assertTrue(_given_error["enabled"] == _read_error["enabled"])
        self.assertTrue(np.allclose(_given_error["err"].error, _read_error["err"].error))
        self.assertTrue(_given_error["err"].corr_coeff == _read_error["err"].corr_coeff)


TEST_PARAMETRIC_MODEL_XY = """
type: xy
x_data: [0.0, 1.0, 2.0, 3.0, 4.0, 5.0]
model_function:
    type: xy
    python_code: |
        def linear_model(x, a, b):
            return a * x + b
model_parameters: [1.1, -1.5]
"""

TEST_PARAMETRIC_MODEL_XY_MISSING_KEYWORD = """
type: xy
model_function:
    type: xy
    python_code: |
        def linear_model(x, a, b):
            return a * x + b
model_parameters: [1.1, -1.5]
"""

TEST_PARAMETRIC_MODEL_XY_EXTRA_KEYWORD = (
    TEST_PARAMETRIC_MODEL_XY
    + """
extra_keyword: 3.14
"""
)

TEST_PARAMETRIC_MODEL_XY_WITH_ERRORS = (
    TEST_PARAMETRIC_MODEL_XY
    + """
x_errors:
  - correlation_coefficient: 0.0
    error_value: 0.1
    name: test_x_error
    relative: false
    type: simple
"""
)


class TestXYParametricModelYamlRepresenter(unittest.TestCase):
    @staticmethod
    def linear_model(x, a, b):
        return a * x + b

    def setUp(self):
        self._test_x = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        self._test_model_parameters = np.array([1.1, -1.5])
        self._test_parametric_model = XYParametricModel(
            self._test_x,
            ModelFunctionBase(TestXYParametricModelYamlRepresenter.linear_model),
            self._test_model_parameters,
        )

        self._test_parametric_model_with_errors = XYParametricModel(
            self._test_x,
            ModelFunctionBase(TestXYParametricModelYamlRepresenter.linear_model),
            self._test_model_parameters,
        )
        self._test_parametric_model_with_errors.add_error(axis="x", err_val=0.1, name="test_x_error", correlation=0, relative=False)

        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._roundtrip_stringstream_with_errors = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_XY))
        self._testfile_stringstream_with_errors = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_XY_WITH_ERRORS))

        self._roundtrip_streamreader = ParametricModelYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamreader_with_errors = ParametricModelYamlReader(self._roundtrip_stringstream_with_errors)
        self._roundtrip_streamwriter = ParametricModelYamlWriter(self._test_parametric_model, self._roundtrip_stringstream)
        self._roundtrip_streamwriter_with_errors = ParametricModelYamlWriter(
            self._test_parametric_model_with_errors, self._roundtrip_stringstream_with_errors
        )
        self._testfile_streamreader = ParametricModelYamlReader(self._testfile_stringstream)
        self._testfile_streamreader_with_errors = ParametricModelYamlReader(self._testfile_stringstream_with_errors)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_XY_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_XY_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ParametricModelYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ParametricModelYamlReader(self._testfile_stringstream_extra_keyword)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_write_to_roundtrip_stringstream_with_errors(self):
        self._roundtrip_streamwriter_with_errors.write()

    def test_read_from_testfile_stream(self):
        _read_parametric_model = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_parametric_model, XYParametricModel))

        self.assertTrue(np.allclose(_read_parametric_model.parameters, self._test_model_parameters))
        _read_parametric_model.x = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        self.assertTrue(np.allclose(_read_parametric_model.y, self._test_parametric_model.y))

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_read_from_testfile_stream_with_errors(self):
        _read_parametric_model = self._testfile_streamreader_with_errors.read()
        self.assertTrue(isinstance(_read_parametric_model, XYParametricModel))

        self.assertTrue(np.allclose(_read_parametric_model.parameters, self._test_model_parameters))
        _read_parametric_model.x = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        self.assertTrue(np.allclose(_read_parametric_model.y, self._test_parametric_model_with_errors.y))

        _given_error = self._test_parametric_model_with_errors.get_error("test_x_error")
        _read_error = _read_parametric_model.get_error("test_x_error")
        self.assertTrue(_given_error["axis"] == _read_error["axis"])
        self.assertTrue(_given_error["enabled"] == _read_error["enabled"])
        self.assertTrue(np.allclose(_given_error["err"].error, _read_error["err"].error))
        self.assertTrue(_given_error["err"].corr_coeff == _read_error["err"].corr_coeff)

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_parametric_model = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_parametric_model, XYParametricModel))

        self.assertTrue(np.allclose(_read_parametric_model.parameters, self._test_model_parameters))
        _read_parametric_model.x = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        self.assertTrue(np.allclose(_read_parametric_model.y, self._test_parametric_model.y))

    def test_round_trip_with_stringstream_with_errors(self):
        self._roundtrip_streamwriter_with_errors.write()
        self._roundtrip_stringstream_with_errors.seek(0)  # return to beginning
        _read_parametric_model = self._roundtrip_streamreader_with_errors.read()
        self.assertTrue(isinstance(_read_parametric_model, XYParametricModel))

        self.assertTrue(np.allclose(_read_parametric_model.parameters, self._test_model_parameters))
        _read_parametric_model.x = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        self.assertTrue(np.allclose(_read_parametric_model.y, self._test_parametric_model_with_errors.y))

        _given_error = self._test_parametric_model_with_errors.get_error("test_x_error")
        _read_error = _read_parametric_model.get_error("test_x_error")
        self.assertTrue(_given_error["axis"] == _read_error["axis"])
        self.assertTrue(_given_error["enabled"] == _read_error["enabled"])
        self.assertTrue(np.allclose(_given_error["err"].error, _read_error["err"].error))
        self.assertTrue(_given_error["err"].corr_coeff == _read_error["err"].corr_coeff)



# flake8: noqa F401 (imported but not used)
import unittest

import numpy as np
import six

from kafe2.core.fitters.nexus import Function, Nexus, Parameter
from kafe2.core.fitters.nexus_fitter import NexusFitter
from kafe2.core.minimizers import AVAILABLE_MINIMIZERS


class AbstractTestNexusFitter(object):
    MINIMIZER = None

    @staticmethod
    def slsq(x, y, x_1, y_1, x_2, y_2):
        return (x - x_1) ** 2 + (y - y_1) ** 2 + (x - x_2) ** 2 + (y - y_2) ** 2

    def setUp(self):
        self.nexus = Nexus()

        self._ref_xy_1 = (3, 9)
        self._ref_xy_2 = (7, 11)
        self._ref_xy = tuple(0.5 * (np.array(self._ref_xy_1) + np.array(self._ref_xy_2)))
        self._startval_x = 300
        self._startval_y = -20

        par_x_1, par_y_1 = (
            self.nexus.add(Parameter(self._ref_xy_1[0], name="x_1")),
            self.nexus.add(Parameter(self._ref_xy_1[1], name="y_1")),
        )
        par_x_2, par_y_2 = (
            self.nexus.add(Parameter(self._ref_xy_2[0], name="x_2")),
            self.nexus.add(Parameter(self._ref_xy_2[1], name="y_2")),
        )

        par_x, par_y = (
            self.nexus.add(Parameter(self._startval_x, name="x")),
            self.nexus.add(Parameter(self._startval_y, name="y")),
        )

        self.nexus.add_function(self.slsq)

        self.fitter = NexusFitter(
            self.nexus,
            parameters_to_fit=("x", "y"),
            parameter_to_minimize="slsq",
            minimizer=self.MINIMIZER,
        )

    def _assert_fit_results(self, xy_val=None):
        xy_val = xy_val or self._ref_xy

        for _par_name, _val in zip(("x", "y"), xy_val):
            self.assertAlmostEqual(self.fitter.get_fit_parameter_values()[_par_name], _val, places=3)

        _slsq_args = tuple(xy_val) + self._ref_xy_1 + self._ref_xy_2
        self.assertAlmostEqual(self.fitter.parameter_to_minimize_value, self.slsq(*_slsq_args), places=2)

    # -- init

    def test_init_inexistent_fit_parameters_raise(self):
        with self.assertRaises(ValueError):
            NexusFitter(
                self.nexus,
                parameters_to_fit=("bogus_parameter",),
                parameter_to_minimize="slsq",
            )

    def test_init_inexistent_min_parameter_raise(self):
        with self.assertRaises(ValueError):
            NexusFitter(
                self.nexus,
                parameters_to_fit=("x", "y"),
                parameter_to_minimize="bogus_parameter",
            )

    # -- basic interface

    def test_simple_properties(self):
        self.assertEqual(self.fitter.parameters_to_fit, ("x", "y"))
        self.assertEqual(self.fitter.parameter_to_minimize, "slsq")
        self.assertEqual(
            self.fitter.get_fit_parameter_values(),
            {
                "x": self._startval_x,
                "y": self._startval_y,
            },
        )
        self.assertEqual(self.fitter.n_fit_par, 2)
        self._assert_fit_results(xy_val=[self._startval_x, self._startval_y])

    def test_set_all_fit_parameter_values(self):
        self.fitter.set_all_fit_parameter_values([7, 2])
        self._assert_fit_results(xy_val=[7, 2])

    def test_set_all_fit_parameter_values_length_mismatch_raise(self):
        with self.assertRaises(ValueError):
            self.fitter.set_all_fit_parameter_values([7, 2, 30])
        with self.assertRaises(ValueError):
            self.fitter.set_all_fit_parameter_values([2])

    def test_set_fit_parameter_values(self):
        self.fitter.set_fit_parameter_values(x=7, y=2)
        self._assert_fit_results(xy_val=[7, 2])

    def test_set_fit_parameter_values_unknown_raise(self):
        with self.assertRaises(ValueError):
            self.fitter.set_fit_parameter_values(bogus_parameter=7)

    def test_do_fit(self):
        self.fitter.do_fit()
        self._assert_fit_results()  # nominal fit results

    def test_state_is_from_minimizer(self):
        self.assertEqual(self.fitter.state_is_from_minimizer, False)
        self.fitter.do_fit()
        self.assertEqual(self.fitter.state_is_from_minimizer, True)
        self.fitter.set_fit_parameter_values(x=7)
        self.assertEqual(self.fitter.state_is_from_minimizer, False)

    # -- parameter errors

    def test_fit_parameter_errors(self):
        self.fitter.do_fit()
        self.assertAlmostEqual(self.fitter.fit_parameter_errors[0], 1.0 / np.sqrt(2))
        self.assertAlmostEqual(self.fitter.fit_parameter_errors[1], 1.0 / np.sqrt(2))

    def test_fit_parameter_cov_mat(self):
        self.fitter.do_fit()
        self.assertAlmostEqual(self.fitter.fit_parameter_cov_mat[0, 0], 0.5)

    def test_fit_parameter_cor_mat(self):
        self.fitter.do_fit()
        self.assertAlmostEqual(self.fitter.fit_parameter_cor_mat[0, 0], 1.0)

    def test_asymmetric_fit_parameter_errors(self):
        self.fitter.do_fit()
        self.assertAlmostEqual(self.fitter.asymmetric_fit_parameter_errors[0][0], -1.0 / np.sqrt(2), places=4)
        self.assertAlmostEqual(self.fitter.asymmetric_fit_parameter_errors[0][1], 1.0 / np.sqrt(2), places=4)

    def test_asymmetric_fit_parameter_errors_if_calculated(self):
        self.fitter.do_fit()
        self.assertIs(self.fitter.asymmetric_fit_parameter_errors_if_calculated, None)

    # -- fixing & limiting parameters

    def test_fix_parameter(self):
        self.fitter.fix_parameter("x")
        self.fitter.do_fit()
        self._assert_fit_results(xy_val=[self._startval_x, self._ref_xy[1]])

    def test_fix_parameter_to_value(self):
        self.fitter.fix_parameter("x", 200)
        self.fitter.do_fit()
        self._assert_fit_results(xy_val=[200, self._ref_xy[1]])

    def test_release_parameter(self):
        self.fitter.fix_parameter("x")
        self.fitter.do_fit()
        self.fitter.release_parameter("x")
        self.fitter.do_fit()
        self._assert_fit_results()  # nominal fit results

    def test_limit_parameter(self):
        self.fitter.limit_parameter("x", (6, 10))
        self.fitter.do_fit()
        self._assert_fit_results(xy_val=[6, self._ref_xy[1]])

    def test_unlimit_parameter(self):
        self.fitter.limit_parameter("x", (6, 10))
        self.fitter.do_fit()
        self.fitter.unlimit_parameter("x")
        self.fitter.do_fit()
        self._assert_fit_results()  # nominal fit results

    # -- profiles & contours

    def test_profile_without_do_fit(self):
        with self.assertRaises(RuntimeError):
            self.fitter.profile("x")

    def test_profile(self):
        self.fitter.do_fit()
        p = self.fitter.profile("x", size=20, sigma=2)[0]
        self.assertEqual(p.shape, (2, 20))
        self.assertAlmostEqual(p[0][-1] - p[0][0], 4 * self.fitter.fit_parameter_errors[0])

    def test_contour_without_do_fit(self):
        with self.assertRaises(RuntimeError):
            self.fitter.contour("x", "y")

    def test_contour(self):
        self.fitter.do_fit()
        c = self.fitter.contour("x", "y")
        # TODO: test result?


if "scipy" in AVAILABLE_MINIMIZERS:

    class TestNexusFitterScipy(AbstractTestNexusFitter, unittest.TestCase):
        MINIMIZER = "scipy"


if "iminuit" in AVAILABLE_MINIMIZERS:

    class TestNexusFitterIMinuit(AbstractTestNexusFitter, unittest.TestCase):
        MINIMIZER = "iminuit"



import abc

import six

# import fit classes
from ...custom import CustomFit
from ...histogram import HistFit
from ...indexed import IndexedFit
from ...unbinned import UnbinnedFit
from ...xy import XYFit
from .._base import GenericDReprBase

__all__ = ["FitDReprBase"]


@six.add_metaclass(abc.ABCMeta)
class FitDReprBase(GenericDReprBase):
    BASE_OBJECT_TYPE_NAME = "fit"

    _CLASS_TO_OBJECT_TYPE_NAME = {
        CustomFit: "custom",
        HistFit: "histogram",
        IndexedFit: "indexed",
        UnbinnedFit: "unbinned",
        XYFit: "xy",
    }
    _OBJECT_TYPE_NAME_TO_CLASS = {
        "custom": CustomFit,
        "histogram": HistFit,
        "indexed": IndexedFit,
        "unbinned": UnbinnedFit,
        "xy": XYFit,
    }

    def __init__(self, fit=None):
        self._kafe_object = fit
        super(FitDReprBase, self).__init__()



import sys
import warnings

from ...config import kc

__all__ = ["get_minimizer"]

AVAILABLE_MINIMIZERS = dict()

_MINIMIZER_NAME_ALIASES = dict()

_catch_error_class = ImportError
if sys.version_info >= (3, 6):
    # python version 3.6+ throws a different exception type on import fail...
    _catch_error_class = ModuleNotFoundError

try:
    from .scipy_optimize_minimizer import MinimizerScipyOptimize

    __all__.append("MinimizerScipyOptimize")
    AVAILABLE_MINIMIZERS.update(
        {
            "scipy": MinimizerScipyOptimize,
        }
    )
    _MINIMIZER_NAME_ALIASES["scipy.optimize"] = "scipy"
except _catch_error_class:
    pass

try:
    from .iminuit_minimizer import MinimizerIMinuit

    __all__.append("MinimizerIMinuit")
    AVAILABLE_MINIMIZERS.update(
        {
            "iminuit": MinimizerIMinuit,
        }
    )
except _catch_error_class:
    warnings.warn("iminuit is not installed. While not strictly needed its use is " "heavily recommended for better performance and convergence.")
except SyntaxError:  # Newer versions of iminuit do not support Python 2.
    pass

try:
    from .root_tminuit_minimizer import MinimizerROOTTMinuit

    __all__.append("MinimizerROOTTMinuit")
    AVAILABLE_MINIMIZERS.update(
        {
            "root.tminuit": MinimizerROOTTMinuit,
        }
    )
    _MINIMIZER_NAME_ALIASES["minuit"] = "root.tminuit"
    _MINIMIZER_NAME_ALIASES["root"] = "root.tminuit"
except (
    _catch_error_class,
    ImportError,
):  # Python 2.7 ROOT bindings cause ImportError in Python 3.6
    pass

# raise if no minimizers can be imported
if not AVAILABLE_MINIMIZERS:
    raise RuntimeError(
        "Fatal error: no minimizers found! Please ensure that "
        "at least one of the following Python packages is installed: "
        "['scipy', 'iminuit', 'ROOT']"
    )


def get_minimizer(minimizer_spec=None):
    """Creates a MinimizerBase object from a given minimizer name.

    :param minimizer_spec: Name of the minimizer to return.
    :type minimizer_spec: str or None
    :return: MinimizerBase-derived kafe2 minimizer object
    """
    global AVAILABLE_MINIMIZERS
    # for 'None', return the default minimizer
    if minimizer_spec is None:
        # go through the default minimizers in the order specified in config
        _minimizer_specs = kc("core", "minimizers", "default_minimizer_list")

        # try every spec until a minimizer is found
        for _minimizer_spec in _minimizer_specs:
            _minimizer_spec = _minimizer_spec.lower()
            _minimizer_spec = _MINIMIZER_NAME_ALIASES.get(_minimizer_spec, _minimizer_spec)
            _minimizer = AVAILABLE_MINIMIZERS.get(_minimizer_spec, None)
            if _minimizer is not None:
                return _minimizer

        raise ValueError(
            "Could not find any minimizer in default list: {}! " "Available: {}".format(_minimizer_specs, list(AVAILABLE_MINIMIZERS.keys()))
        )
    _minimizer_spec = minimizer_spec.lower()
    _minimizer_spec = _MINIMIZER_NAME_ALIASES.get(_minimizer_spec, _minimizer_spec)
    _minimizer = AVAILABLE_MINIMIZERS.get(_minimizer_spec, None)
    if _minimizer is not None:
        return _minimizer

    raise ValueError("Unknown minimizer '{}'! " "Available: {}".format(minimizer_spec, list(AVAILABLE_MINIMIZERS.keys())))



from .._base.cost import (
    CostFunction,
    CostFunction_Chi2,
    CostFunction_GaussApproximation,
    CostFunction_NegLogLikelihood,
)

__all__ = [
    "HistCostFunction",
    "HistCostFunction_Chi2",
    "HistCostFunction_NegLogLikelihood",
    "HistCostFunction_GaussApproximation",
]


class HistCostFunction(CostFunction):
    pass


class HistCostFunction_Chi2(CostFunction_Chi2):
    pass


class HistCostFunction_NegLogLikelihood(CostFunction_NegLogLikelihood):
    pass


class HistCostFunction_GaussApproximation(CostFunction_GaussApproximation):
    pass



from copy import deepcopy

from .._base import DataContainerBase, FitBase
from .._base.cost import CostFunction_NegLogLikelihood
from ..util import (  # noqa: F401 (collect imported but not used)
    collect,
    function_library,
)
from .container import HistContainer
from .model import HistModelFunction, HistParametricModel
from .plot import HistPlotAdapter

__all__ = ["HistFit"]


class HistFit(FitBase):
    CONTAINER_TYPE = HistContainer
    MODEL_TYPE = HistParametricModel
    MODEL_FUNCTION_TYPE = HistModelFunction
    PLOT_ADAPTER_TYPE = HistPlotAdapter
    RESERVED_NODE_NAMES = {
        "data",
        "model",
        "model_density",
        "cost",
        "data_error",
        "model_error",
        "total_error",
        "data_cov_mat",
        "model_cov_mat",
        "total_cov_mat",
        "data_cor_mat",
        "model_cor_mat",
        "total_cor_mat",
    }
    _BASIC_ERROR_NAMES = {"data_error", "model_error", "data_cov_mat", "model_cov_mat"}

    def __init__(
        self,
        data,
        model_function=function_library.normal_distribution,
        cost_function=CostFunction_NegLogLikelihood(data_point_distribution="poisson"),
        bin_evaluation="simpson",
        density=True,
        minimizer=None,
        minimizer_kwargs=None,
        dynamic_error_algorithm="nonlinear",
    ):
        """
        Construct a fit of a model to a histogram. If bin_evaluation is a Python function or
        of a numpy.vectorize object it is interpreted as the antiderivative of
        model_density_function. If bin_evaluation is equal to "rectangle", "midpoint", "trapezoid",
        or "simpson" the bin heights are evaluated according to the corresponding quadrature
        formula. If bin_evaluation is equal to "numerical" the bin heights are evaluated by
        numerically integrating model_density_function.

        :param data: an encapsulated representation of the histogrammed data.
        :type data: :py:class:`~kafe2.fit.hist.HistContainer` or two-dimensional iterable of bin
            heights and bin edges as returned by `np.histogram`.
        :param model_function: the model (density) function
        :type model__function: :py:class:`~kafe2.fit.hist.HistModelFunction` or unwrapped
            native Python function
        :param cost_function: the cost function
        :type cost_function: :py:class:`~kafe2.fit._base.CostFunctionBase`-derived or unwrapped
            native Python function
        :param bin_evaluation: how the model evaluates bin heights.
        :type bin_evaluation: str, callable, or numpy.vectorize
        :param density: if True, scale model function to the number of data points.
        :type density: bool
        :param minimizer: the minimizer to use for fitting.
        :type minimizer: None, "iminuit", "tminuit", or "scipy".
        :param minimizer_kwargs: dictionary with kwargs for the minimizer.
        :type minimizer_kwargs: dict
        """
        self._bin_evaluation = bin_evaluation
        self._density = density
        super(HistFit, self).__init__(
            data=data,
            model_function=model_function,
            cost_function=cost_function,
            minimizer=minimizer,
            minimizer_kwargs=minimizer_kwargs,
            dynamic_error_algorithm=dynamic_error_algorithm,
        )

    # -- private methods

    def _init_nexus(self):
        super(HistFit, self)._init_nexus()

        self._nexus.add_dependency("model", depends_on=("parameter_values"))

    def _set_new_data(self, new_data):
        try:
            _new_data_is_numpy_histogram = len(new_data) == 2 and len(new_data[0]) + 1 == len(new_data[1])
        except TypeError:
            _new_data_is_numpy_histogram = False
        if _new_data_is_numpy_histogram:
            self._data_container = self.CONTAINER_TYPE(
                n_bins=len(new_data[0]),
                bin_range=(new_data[1][0], new_data[1][-1]),
                bin_edges=new_data[1],
            )
            self._data_container.set_bins(new_data[0])
        elif isinstance(new_data, self.CONTAINER_TYPE):
            self._data_container = deepcopy(new_data)
        elif isinstance(new_data, DataContainerBase):
            raise TypeError("Incompatible container type '{}' (expected '{}')".format(type(new_data), self.CONTAINER_TYPE))
        else:
            raise TypeError("Fitting a histogram requires a kafe2 HistContainer or a NumPy histogram as data " f"but received {new_data}")
        self._data_container._on_error_change_callback = self._on_error_change

        self._nexus.get("data").mark_for_update()

    def _set_new_parametric_model(self):
        # create the child ParametricModel object
        self._param_model = HistParametricModel(
            self._data_container.size,
            self._data_container.bin_range,
            self._model_function,
            self.parameter_values,
            self._data_container.bin_edges,
            bin_evaluation=self._bin_evaluation,
            density=self._density,
        )

    # -- public properties

    @property
    def model(self):
        """array of model predictions for the data points"""
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        # if model is just a density, scale it to the number of data points
        if self._param_model.density:
            return self._param_model.data * self._data_container.n_entries
        else:
            return self._param_model.data

    @property
    def density(self):
        return self._param_model.density

    # FIXME: how to handle scaling of model_error

    # -- public methods

    # add_error... methods inherited from FitBase ##

    def eval_model_function_density(self, x, model_parameters=None):
        """
        Evaluate the model function density.

        :param x: values of *x* at which to evaluate the model function density
        :type x: iterable of float
        :param model_parameters: the model parameter values (if ``None``, the current values are used)
        :type model_parameters: iterable of float
        :return: model function density values
        :rtype: :py:class:`numpy.ndarray`
        """
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        return self._param_model.eval_model_function_density(x=x, model_parameters=model_parameters)



r"""The :py:mod:`kafe2.fit` module provides an object-oriented toolkit for estimating model
parameters from data ("fitting").

It distinguishes between a number of different data types:

*  *xy* data (dedicated submodule: :py:mod:`~kafe2.fit.xy`),
*  series of indexed measurements (dedicated submodule: :py:mod:`~kafe2.fit.indexed`),
*  histograms (dedicated submodule: :py:mod:`~kafe2.fit.histogram`),
*  raw 1D data using the method of maximum likelihood ("unbinned fit",
   dedicated submodule: :py:mod:`~kafe2.fit.histogram`), and
*  direct minimization of a cost function (dedicated submodule: :py:mod:`~kafe2.fit.custom`).

Each of the above data types has its own particularities when it comes to fitting.
The main difference is due to the way uncertainties can be defined and
interpreted for each type of data and how the fit results are presented.


XY Data
-------

For *xy* data, one data set consists of a list of :math:`N` distinct :math:`y` measurements
:math:`d_i` with the (discrete) index :math:`i` ranging from :math:`0` to :math:`N-1`.
The measurements were taken at :math:`x` values :math:`x_i`.
For each measurement in the series, one or more uncertainty sources can be defined,
each being a numerical estimate of how much the respective measurement has fluctuated from the
"true values".
Correlations between uncertainties on separate measurements :math:`d_i` and :math:`d_j` can also
be taken into account by using *covariance/correlation matrices*.

Additional uncertainites on :math:`x_i` can also be defined.
When fitting an *xy* model to data they are converted to :math:`y` uncertainties via multiplication
with the derivative of the model function by :math:`x`.
When plotting the result of *xy* fits, the model function is displayed as a continuous
function of :math:`x`, and an *error band* can be computed to reflect the model uncertainty,
as determined by propagating the parameter uncertainties onto the *y* axis.

The following objects are provided for handling *xy* data:

*  :py:obj:`~kafe2.fit.xy.XYContainer`: data container for storing *xy* data
*  :py:obj:`~kafe2.fit.xy.XYParametricModel`: corresponding model
*  :py:obj:`~kafe2.fit.xy.XYFit`: a fit of a parametric model to *xy* data

Indexed data
------------

Compared to *xy* data indexed data no longer has an explicit *x* axis.
The data simply appears as an indexed list of data points.
As a consequence the model function does not expect an independent variable.

The following objects are provided for handling indexed data, as described above:

*  :py:obj:`~kafe2.fit.indexed.IndexedContainer`: data container for storing indexed data
*  :py:obj:`~kafe2.fit.indexed.IndexedParametricModel`: corresponding model
*  :py:obj:`~kafe2.fit.indexed.IndexedFit`: a fit of a parametric model to *indexed* data

Histograms
----------

*kafe2* is also able to handle *histograms*. Histograms organize measurements whose
values can fall anywhere across a continuum of values into a number of discrete regions
or "bins". Typically, the continuous "measurement space" (a closed real interval
:math:`[x_{\rm min}, x_{\rm max}]`)
is subdivided into a sequence of successive intervals at the "bin edges"
:math:`x_{\rm min} < x_1 < x_2 < \ldots < x_{\rm max}`. Whenever a measurement falls into one of
the bins, the value of that histogram bin is incremented by one.
A histogram is completely defined by its bin edges and the bin values.

.. note::
    The bin numbering starts at :math:`1` for the first bin and ends at :math:`N`, where :math:`N`
    is defined as the *size* of the histogram. The bin numbers :math:`0` and :math:`N+1` refer to
    the underflow (below :math:`x_{\rm min}`) and overflow bin (above :math:`x_{\rm max}`),
    respectively.


Defining a parametric model for histograms is not as straightforward as for *xy* and indexed data.
Seeing as they keep track of the number of entries in different intervals of the continuum, the bin
values can be interpreted using probability theory.

As the number of entries approaches infinity, the number of entries :math:`n` in the bin covering an
interval :math:`[a, b)`, divided by the total number of entries :math:`N_{\rm E}`, will approach the
probablity of an event landing in that bin:

.. math::
    \lim_{N_{\rm E}\rightarrow\infty} \frac{n}{N_{\rm E}} = \int_a^b f(x)\,{\rm d}x = F(b) - F(a)

In the above formula, :math:`f(x)` is the *probability density function*,
and :math:`F(x)` is an antiderivative of :math:`f`
(for example the *cumulative distribution function*).

Using the above relation, the model prediction :math:`m` for the bin :math:`[a, b)` can be defined
as:

.. math::
    m = N_{\rm E} \int_a^b f(x)\,{\rm d}x = N_{\rm E} \left(F(b) - F(a)\right)

This means that, for histograms, the model density :math:`f(x)` needs to be specified as the model
function.
The model is then calculated by numerically integrating this function over each bin.

An alternative would be to specify the model density *antiderviative* :math:`F` alongside
the model, so that the model can be calculated as a simple difference, rather than as an integral.

The following objects are provided for handling histograms:

*  :py:obj:`~kafe2.fit.histogram.HistContainer`: data container for storing histograms
*  :py:obj:`~kafe2.fit.histogram.HistParametricModel`: corresponding model
*  :py:obj:`~kafe2.fit.histogram.HistFit`: a fit of a parametric model to histograms

Unbinned
--------

If data is treated as *unbinned* the model function :math:`f(x)` is interpreted as a
*model density function*.
The cost function value :math:`C` is then directly calculated as the negative log-likelihood of the
data given said PDF:

.. math::
    C = - 2 \sum_{i=0}^{N-1} \ln f(x_i) .

An unbinned fit is the edge case of a histogram fit for as the individual bins become
infinitessimally thin.

The following objects are provided for handling unbinned data:

*  :py:obj:`~kafe2.fit.unbinned.UnbinnedContainer`: data container for storing unbinned data
*  :py:obj:`~kafe2.fit.unbinned.UnbinnedParametricModel`: corresponding model
*  :py:obj:`~kafe2.fit.unbinned.UnbinnedFit`: a fit of a parametric model to unbinned data

Custom
------

Lets the user directly define a cost function.
Since this fit type does not have explicit data the fit results cannot be plotted automatically.

The following objects are provided for custom fits:

*  :py:obj:`~kafe2.fit.custom.CustomFit`: a fit for minimizing a cost function

Plots
-----

For creating graphical representations of fits, the :py:obj:`~kafe2.fit.Plot` is provided.
It can be instantiated with any fit object (or list of fit objects) as an argument and will produce
one or more plots accordingly using `matplotlib`.

:synopsis: This module contains specialized objects for storing measurement data,
           defining and fitting parametric models to these data and producing
           graphical representations ("plots") of the result.
           It relies on the :py:mod:`kafe2.core` module for basic functionality.

.. moduleauthor:: Daniel Savoiu <daniel.savoiu@cern.ch>
"""

# public interface of submodule 'kafe2.fit'

# WARNING: Changing the order of these imports might break the kafe2 interface
# flake8: noqa F401, F403 (unused import, unable to detect undefined names from '*' import)
# fmt: off
# isort: off
from .custom import *
from .indexed import *
from .histogram import *
from .multi import *
from .tools import *
from .xy import *
from .unbinned import *

from ._base.plot import Plot
from .tools.fit_wrapper import Fit  # import after every fit to avoid import conflicts with other tools

from .util import function_library, custom_fit, hist_fit, indexed_fit, unbinned_fit, xy_fit, plot, \
    k2Fit
# fmt: on
# isort: on



from ....fit import HistContainer, IndexedContainer, UnbinnedContainer, XYContainer
from .. import _AVAILABLE_REPRESENTATIONS
from .._yaml_base import YamlReaderMixin, YamlWriterMixin
from ..error import common_error_tools
from ._base import DataContainerDReprBase

__all__ = ["DataContainerYamlReader", "DataContainerYamlWriter"]


class DataContainerYamlWriter(YamlWriterMixin, DataContainerDReprBase):
    DUMPER = common_error_tools.MatrixYamlDumper

    def __init__(self, data_container, output_io_handle):
        super(DataContainerYamlWriter, self).__init__(output_io_handle=output_io_handle, data_container=data_container)

    @classmethod
    def _make_representation(cls, container):
        """Create a dictionary representing a data container.

        :param container: The data container which will be converted.
        :type container: kafe2.fit._base.DataContainerBase
        """
        _yaml_doc = dict()
        _class = container.__class__

        # -- determine container type
        _type = cls._CLASS_TO_OBJECT_TYPE_NAME.get(_class, None)
        if _type is None:
            raise TypeError("Container type unknown or not supported: %s" % _class)
        _yaml_doc["type"] = _type

        # -- write representation for container types
        if _class is HistContainer:
            _yaml_doc["bin_edges"] = list(map(float, container.bin_edges))
            if container._manual_heights:
                _yaml_doc["bin_heights"] = list(map(float, container.data))  # float64 -> float
                _yaml_doc["underflow"] = float(container.underflow)
                _yaml_doc["overflow"] = float(container.underflow)
            else:
                _yaml_doc["raw_data"] = list(map(float, container.raw_data))  # float64 -> float
        elif _class is IndexedContainer or _class is UnbinnedContainer:
            _yaml_doc["data"] = container.data.tolist()
        elif _class is XYContainer:
            _yaml_doc["x_data"] = container.x.tolist()
            _yaml_doc["y_data"] = container.y.tolist()
        else:
            raise TypeError("Container type unknown or not supported: {}".format(_type))

        # -- write error representation for all container types
        if container.has_errors:
            common_error_tools.write_errors_to_yaml(container, _yaml_doc)

        # write labels for all container types
        if container.label is not None:
            _yaml_doc["label"] = container.label
        _x_axis_label, _y_axis_label = container.axis_labels
        if _x_axis_label is not None:
            _yaml_doc["x_label"] = _x_axis_label
        if _y_axis_label is not None:
            _yaml_doc["y_label"] = _y_axis_label

        return _yaml_doc


class DataContainerYamlReader(YamlReaderMixin, DataContainerDReprBase):
    LOADER = common_error_tools.MatrixYamlLoader

    def __init__(self, input_io_handle):
        super(DataContainerYamlReader, self).__init__(input_io_handle=input_io_handle, data_container=None)

    @classmethod
    def _get_required_keywords(cls, yaml_doc, container_class):
        if container_class is HistContainer:
            return []
        if container_class is IndexedContainer or container_class is UnbinnedContainer:
            return ["data"]
        if container_class is XYContainer:
            return ["x_data", "y_data"]
        raise TypeError("Unknown container type")

    @classmethod
    def _convert_yaml_doc_to_object(cls, yaml_doc):
        # -- determine container class from type
        _container_type = yaml_doc.pop("type")
        _class = cls._OBJECT_TYPE_NAME_TO_CLASS.get(_container_type)

        # -- read in representation for container types
        if _class is HistContainer:
            _bin_edges = yaml_doc.pop("bin_edges", None)
            _n_bins = yaml_doc.pop("n_bins", None)
            _bin_range = yaml_doc.pop("bin_range", None)
            if not _bin_edges and not (_n_bins and _bin_range):
                raise ValueError("When reading in a histogram dataset either " "bin_edges or n_bins and bin_range has to be specified!")
            if _bin_edges:
                _n_bins = len(_bin_edges) - 1
                _bin_range = (_bin_edges[0], _bin_edges[-1])
            _raw_data = yaml_doc.pop("raw_data", None)
            _bin_heights = yaml_doc.pop("bin_heights", None)
            if _raw_data and _bin_heights:
                raise ValueError("When reading in a histogram dataset only one out of " "raw_data and bin_heights can be specified!")
            _container_obj = HistContainer(n_bins=_n_bins, bin_range=_bin_range, bin_edges=_bin_edges, fill_data=_raw_data)
            if _bin_heights:
                _underflow = yaml_doc.pop("underflow", 0)
                _overflow = yaml_doc.pop("overflow", 0)
                _container_obj.set_bins(bin_heights=_bin_heights, underflow=_underflow, overflow=_overflow)
        elif _class is IndexedContainer:
            _data = yaml_doc.pop("data")
            _container_obj = IndexedContainer(_data)
        elif _class is UnbinnedContainer:
            _data = yaml_doc.pop("data")
            _container_obj = UnbinnedContainer(_data)
        elif _class is XYContainer:
            _x_data = yaml_doc.pop("x_data")
            _y_data = yaml_doc.pop("y_data")
            _container_obj = XYContainer(_x_data, _y_data)
        else:
            raise TypeError("Container type unknown or not supported: {}".format(_container_type))

        # get labels for all container types
        _container_obj.label = yaml_doc.pop("label", None)
        _container_obj.axis_labels = (yaml_doc.pop("x_label", None), yaml_doc.pop("y_label", None))

        # -- process error sources
        _container_obj, yaml_doc = common_error_tools.process_error_sources(container_obj=_container_obj, yaml_doc=yaml_doc)

        return _container_obj, yaml_doc


# register the above classes in the module-level dictionary
DataContainerYamlReader._register_class(_AVAILABLE_REPRESENTATIONS)
DataContainerYamlWriter._register_class(_AVAILABLE_REPRESENTATIONS)



"""This submodule contains objects for handling different data representations. Whole fit objects can be saved and
loaded to and from a YAML file. Support for more file formats might follow in future updates.

:synopsis: This submodule contains objects for handling different data representations.

.. moduleauthor:: Daniel Savoiu <daniel.savoiu@cern.ch>
"""

# this module-level dict is populated by importing the
# individual modules below
_AVAILABLE_REPRESENTATIONS = dict()

# register different aliases (i.e. file extensions) for file formats
_FILE_FORMAT_ALIAS_RESOLVDICT = dict(yml="yaml")

# flake8: noqa F401, F403 (imported but unused, used but unable to detect undefined names)

from .constraint import *
from .container import *
from .fit import *
from .format import *
from .model import *


def _get_representer(object_type_name, role, file_format):
    _d_roles = _AVAILABLE_REPRESENTATIONS.get(object_type_name, None)
    if _d_roles is None:
        raise TypeError(
            "No representers found for object type '{}'! Available object types "
            "are: {}".format(object_type_name, _AVAILABLE_REPRESENTATIONS.keys())
        )

    _d_fileformats = _d_roles.get(role, None)
    if _d_fileformats is None:
        raise TypeError(
            "No representers with role '{}' found for object type '{}'! " "Available roles are: {}".format(role, object_type_name, _d_roles.keys())
        )

    _file_format_resolve_alias = _FILE_FORMAT_ALIAS_RESOLVDICT.get(file_format, file_format)
    _representer = _d_fileformats.get(_file_format_resolve_alias, None)
    if _representer is None:
        raise TypeError(
            "No representer with role '{}' found for of object type '{}' and file format '{}'! "
            "Available file formats are: {}".format(file_format, role, object_type_name, _d_fileformats.keys())
        )

    return _representer


# -- convenience functions


def get_reader(object_type_name, file_format):
    return _get_representer(object_type_name=object_type_name, role="reader", file_format=file_format)


def get_writer(object_type_name, file_format):
    return _get_representer(object_type_name=object_type_name, role="writer", file_format=file_format)



import abc

import six

# import ModelFunction classes
from ..._base import ModelFunctionBase
from ...histogram import HistModelFunction, HistParametricModel
from ...indexed import IndexedModelFunction, IndexedParametricModel
from ...unbinned import UnbinnedParametricModel
from ...xy import XYParametricModel
from .._base import GenericDReprBase

__all__ = ["ModelFunctionDReprBase", "ParametricModelDReprBase"]


@six.add_metaclass(abc.ABCMeta)
class ModelFunctionDReprBase(GenericDReprBase):
    BASE_OBJECT_TYPE_NAME = "model_function"

    _CLASS_TO_OBJECT_TYPE_NAME = {
        HistModelFunction: "histogram",
        IndexedModelFunction: "indexed",
        ModelFunctionBase: "base",
    }
    _OBJECT_TYPE_NAME_TO_CLASS = {
        "histogram": HistModelFunction,
        "indexed": IndexedModelFunction,
        "unbinned": ModelFunctionBase,
        "xy": ModelFunctionBase,  # type from fit is passed to model function, needs to be resolved
        "base": ModelFunctionBase,
    }

    def __init__(self, model_function=None):
        self._kafe_object = model_function
        super(ModelFunctionDReprBase, self).__init__()


@six.add_metaclass(abc.ABCMeta)
class ParametricModelDReprBase(GenericDReprBase):
    BASE_OBJECT_TYPE_NAME = "model"

    _CLASS_TO_OBJECT_TYPE_NAME = {
        HistParametricModel: "histogram",
        IndexedParametricModel: "indexed",
        UnbinnedParametricModel: "unbinned",
        XYParametricModel: "xy",
    }
    _OBJECT_TYPE_NAME_TO_CLASS = {
        "histogram": HistParametricModel,
        "indexed": IndexedParametricModel,
        "unbinned": UnbinnedParametricModel,
        "xy": XYParametricModel,
    }

    def __init__(self, parametric_model=None):
        self._kafe_object = parametric_model
        super(ParametricModelDReprBase, self).__init__()



"""This submodule contains objects for handling different representations of data containers.

:synopsis: This submodule contains objects for handling different representations of data containers.

.. moduleauthor:: Daniel Savoiu <daniel.savoiu@cern.ch>
"""

# flake8: noqa F401, F403 (imported but unused, used but unable to detect undefined names)

from ._base import *
from .yaml_drepr import *
